from nx5d.repo.base import DataScanBase
import urllib
from nx5d.recipe import *
from nx5d.recipe import my_callable

def test_file_loader():
    c = load_recipe(f'file://{__file__}#my_callable')
    assert c(None, None) == my_callable(None, None)


def test_pymod_loader():
    c = load_recipe(f'pymod:///nx5d.recipe#my_callable')
    assert c(None, None) == my_callable(None, None)



import time
import numpy as np
import xarray as xr

import itertools

class DatasetFactory:
    '''
    Helper class to ingest data into an `xarray.Dataset`
    '''
    
    def __init__(self, **instr):
        '''
        Args:
            **instr: tag -> dict(path=..., chunk=..., dims=...)
              Where the keys of the instrument setup are as documented
              in `.add_instrument()`.
        '''
        self._instr = instr


    def add_instrument(self, tag, pathfmt, chunk=None, dims=None):
        '''
        Args:
            tag: the name of the data array in the resulting `Dataset`
            path: the location of the data array inside the container
              we're reading from. THIS IS A FORMAT, i.e. you can use
              format keys here that will be substituted by the keyword-args
              data supplied to `.__call__()`.
            chunk: either `None`, `"auto"`, `"scan"` or an integer,
              representing the Dask chunking in the 1st (indexing!) dimension.
              Chunking in the other dimension(s) will always be chosed to fit
              the actual data size.
              The chunking value has the following meaning:
              - 0: no chunking, load all the data at once
              - N: where N is an integer. loads N datapoints per chunk
              - `None` or `"auto"`: perform auto-chunking, i.e. load
                as many data points as we can while still staying below
                a certain threshold (typically in the order of 1M datapoints)
              - `"scan"`: select chunk size based on the size of the first
                scan in the series (see `.__call__()`). If the loading operation
                is performed on a single scan, this is the same as `0`. Otherwise
                this will put each scan into its own chunk, assuming that all
                scans are of the same length of points.
            dims: Name of the dimensions that are NOT the index (i.e. first)
              dimension.
        '''
        self._instr[tag] = {
            'path': pathfmt,
            'chunk': chunk,
            'dims': dims
        }


    def __call__(self, h5like, mode: str = None, **fmtkeys):

        if mode in ('series',):
            return self._build_series(h5like, **fmtkeys)

        elif mode in ('single',):
            return self._build_single(h5like, **fmtkeys)

        elif mode in ('auto', None): # auto-detecting?
        
            for k,v in fmtkeys.items():
                if isinstance(v, list) or isinstance(v, tuple):
                    return self._build_series(h5like, **fmtkeys)
            return self._build_single(h5like, **fmtkeys)


    def _build_series(self, h5like, **fmtkeys):
        
        ## Need to make sure all fmtkeys.values() are lists

        collection = lambda x: x if isinstance(x, list) or isinstance(x, tuple) \
            else tuple([x])
        
        series = {
            k:collection(v) for k,v in fmtkeys.items()
        }
        
        combinations = [
            dict([
                j for j in zip(series.keys(), single)
            ]) \
            for single in itertools.product(*series.values())
        ]

        dset = None
        for single in combinations:
            dnew = self._build_single(h5like, **single)
            #print('dset', dset)
            #print('dnew', dnew)
            if dset is None:
                dset = dnew
            else:
                dset = xr.concat((dset, dnew), dim='index', join='override')

        return dset



    def _build_single(self, h5like, **fmtkeys):

        index_size = None

        ds = xr.Dataset()

        for tag, dset_spec in self._instr.items():
            if isinstance(dset_spec, dict):
                dset = dset_spec['path'].format(**fmtkeys)
                chunks = dset_spec.get('chunk', None)
                dims = dset_spec.get('dims', None)
            else:
                dset = dset_spec.format(**fmtkeys)
                chunks = None
                dims = None

            try:
                dnode = h5like[f'{dset}']
                
                if dims is None and len(dnode.shape)>0:
                    dims = ['index'] + [f'{tag}_{i}' for i in range(1, len(dnode.shape)) ]

                data_size = np.prod(dnode.shape)
                #print(f'Loading: {tag} <- {dset}, shape {dnode.shape}, {data_size}')
            except KeyError as e:
                #print(f'Skipping: {tag} ({dset} not found): {e}')
                continue


            # We expect every value to have an `index_size` sized 1st dimension.
            # As a courtesy, we _also_ accept scalars (shape==() or shape==(1,)),
            # but (maybe?) scale scalars up to (index_size,) arrays of the same value.
            if (len(dnode.shape) == 0) or \
               (len(dnode.shape) == 1 and dnode.shape[0] == 1):
                #if index_size is None:
                #    raise RuntimeError(f'msg="Cannot determine realistic resize target '
                #                       f'of scalar-like value" shape={dnode.shape} '
                #                       f'tag={tag}')

                data = dnode[()]

            # If nodes get too large, or if chunking is explicitly enabled,
            # we store the data as dask arrays
            else:

                # For later use.
                if index_size is None:
                    index_size = dnode.shape[0]

                if (data_size > 1e5 and chunks != 0):
                    if chunks in (None, 'auto'):
                        chunk = int(1e7 / np.prod(dnode.shape[1:]))
                        chunks = (chunk, *(dnode.shape[1:]))
                        #print(f'Auto-chunking: {tag} -> {chunks}')
                    elif chunks in ('scan',):
                        chunk = dnode.shape[0]
                        chunks = (chunk, *(dnode.shape[1:]))
                        #print(f'Auto-chunking: {tag} -> {chunks}')
                    from dask.array import from_array as da_from_array                        
                    data = da_from_array(dnode, chunks=chunks)

                # Regular, unchunked, ndarray data
                else:
                    data = dnode[()]


            if dims is None:
                ds[tag] = data
            else:
                ds[tag] = (dims, data)

        return ds



from nx5d.xrd.base import DatasetFactory as DSBase
from nx5d.repo.base import DataScanBase
from nx5d.xrd.signal import QMapper
from nx5d.repo.filesystem import FilesystemWalker
from nx5d.repo.filesystem import DataRepository as FsDataRepoBase
from nx5d.repo.filesystem import DataProposal as FsDataPropBase

import tifffile, xarray, numpy, logging, os, glob, urllib

logger = logging.getLogger(__name__)

def opcast(v):
    # dirty opportunistic cast
    try:
        return int(v)
    except ValueError:
        try:
            return float(v)
        except ValueError:
            return v

class ScanTypeMismatch(RuntimeError):
    def __init__(self, scan_type):
        super().__init__(f'Scan type mismatch: {scan_type}')

class DatasetFactory(DSBase):
    '''
    DatasetFactory implementation for UDKM "air-gap" based data format.

    This isn't a HDF5-based format, but we'll still reuse the
    base class. The ._build_single() method is reimplemented
    and the 1st parameter (formerly "h5like") is re-interpreted.
    Everything else is reusable as-is from the base class.
    '''

    ## The saving format of UDKM is funny (to put it friendly) and unconsistent.
    ## It makes for a good human-readable format, fit for reverse-engineering
    ## in case of a zombie apocalypse. But it's less than ideal for atomated
    ## processing.
    ##
    ## From bits n pieces of information, here's what we know:
    ## 
    ## - *Generally* (but not always), files begin with one or several line(s)
    ##   of column description, TAB-seprarated; the column names themselves
    ##   contain all kinds of characters (besides TAB), and sometimes supplementary
    ##   information like units in brackets. Example:
    ##   ```
    ##   % Column\tColumnt 2\tColumn [3]\n
    ##   ```
    ##   We'll call  this the "column indicator".
    ##
    ## - *Sometimes* (mostly in "rockingcurves" files), a 2nd line containing even
    ##   more columns are named, which we'll call the "datapoint indicator":
    ##   ```
    ##   % StartTime\tEndTime\tImageName\n
    ##   ```
    ##
    ## - Then the data points are pouring in, two lines for each data point.
    ##   The first line is the timing and image name of the data point,
    ##   TAB-separated. The data points correspond to the "datapoint indicator":
    ##   ```
    ##   % 15:00:09.327\t15:00:12.637\timage00000.tif\n
    ##   ```
    ##   Note that sometime the "datapoint indicator" is missing (i.e. the line
    ##   denoting the columns), but the datapoint indication _data_ is nonetheless
    ##   delivered (e.g. for "RSS"-style "scans*.dat" files).
    ##
    ##   The 2nd line is the data itself, corresponding to the "column indicator"
    ##   (note the lack of a leading `% ` here):
    ##   ```
    ##   1.000000\t-5.000000\t1.000000\t13.220000\t0.144264\t842033.695276\t2165.333333\t0.000000\t0.000000\t...\n
    ##   ```
    ##   Note that the number of data entries per line exceeds the number of columns!
    ##   This is because, usually, the last column is something like a "Curve..." or
    ##   "Intensities...", and is usually just a processed version of the data already
    ##   contained in the .TIFF file (typically integrated along one direction).
    ##
    ##   (Of course we have no way of knowing which "column" is the one with multiple
    ##   values, but going forward we're just going to assume that it's the last one).
    

    ## We're listing all the indicator names here, and (manually) assigning them
    ## python-compatible names. We're using a flat structure here for simplicity.
    ## Implicitly, this assumes that there are no duplicates across different .dat
    ## file types (or, if there are, that they're always going to be translated
    ## the same).
    ## If that's not the case, somebody will need to break this up in a per-file
    ## dictionary :-)
    ##
    _default_indicators = {
        # column indicator names from rockingcurves*.dat:        
        'Loop':                 'loop',
        'Theta / deg.':         'theta',
        '2Theta + Corr / deg.': 'ttheta',
        'Crystal':              'crystal_v',
        'Crystal Calib':        'crystal_calib',
        'Rocking Curve':        'images',
        
        # datapoint indicator names from rockingcurves*.data:
        'StartTime':            'start_time',
        'EndTime':              'end_time',
        'ImageName':            'image_name',
        
        # column indicator names form scans*.dat:
        'Loop_Delay':           'loop_delay',
        'Delay':                'delay',
        'Loop_Theta':           'loop_theta',
        'Theta':                'theta',
        'Crystal [V]':          'crystal_v',
        'Crystal [ph/s]':       'crystal_calib',
        'Pilatus [ph/s]':       'cntrate',
	    'Repeats':              'repeats',
        'ImageNumber':          'image_number',
        'Intensities [ph/s]':   'images',
    }

    # Involuntarily (but usefully :-) the UDKM data are already typed.
    # As a type designator, we're using the presence of specific files.
    _scan_types = {
        'sRSM': [ 'rockingcurves{sfx}.dat' ],
        'tRSS': [ 'scans{sfx}.dat' ],
        'xRSS': [ 'allSteps{sfx}.dat' ]
    }
    
    def __init__(self, indicators=None):
        '''
        Initializes UDKM data retriever.

        Args:
            indicators: Optional dictionary to override default indicators
              (i.e. column names from UDKM .dat files). The keys must be
              verbatim column names -- without the TABs, but with all the
              punctuation, case-sensitive. The values must be python
              symbol names (i.e. preferrably lower-case, using underscore '_').
              The names will be used as data keys in the resulting xarray
              dataset.
        '''
        self.indicators = self._default_indicators.copy()        
        if indicators is not None:
            self.indicators.update(indicators)


    def _get_type(self, src, **keys):
        # Tries to determine the type of scan 'src' from presence or absence
        # of specific files.
        for k, markers in self._scan_types.items():
            for m in markers:
                try:
                    with open(os.path.join(src, m.format(**keys)), 'r') as f:
                        return k
                except FileNotFoundError:
                    pass

        raise RuntimeError(f'msg="uknown type" scan="{src}"')


    def typeof(self, src):
        '''
        Returns the scan type of scan at 'src' (efficiently, tries to open
        as little data as possible).
        '''
        return self._get_type(src, sfx=self._scan_suffix(src))


    def paramsof(self, src):
        '''
        Returns the contents of the params file.
        '''
        import pandas, io
        sfx = self._scan_suffix(src)
        comments = []
        kw = {}
        at_key = None
        params_file = os.path.join(src, f'parameters{sfx}.txt')
        try:
            with open(params_file, 'rb') as f:
                for line in f:
                    try:
                        txt = line.decode('utf-8').strip()
                    except:
                        txt = line.decode('windows-1252').strip()

                    if ': ' in txt:
                        k, v = txt.split(': ')
                        kw[k] = v
                        at_key = k
                    else:
                        if len(txt) == 0:
                            continue
                        if at_key is not None:
                            kw[at_key] += txt
                        else:
                            comments.append(txt)
            kw['Comments'] = ' '.join(comments)
            return kw
        except FileNotFoundError as e:
            logger.warning(e)
            return {}


    def _open_as_type(self, src, scan_type, **keys):
        # Tries to open a specific scan as type 'scan_type'.
        # Raises an error (typically FileNotFound) if file is
        # of a different type.
        # Returns a file handle (open(...)) on success.
        for m in self._scan_types[scan_type]:
            try:
                fname = m.format(**keys)
                #print(f'trying type {scan_type}, signature file {fname}, path {src}')
                fobj = open(os.path.join(src, fname), 'r')
                #print(f'success with {fname}')
                return fobj
            except FileNotFoundError as e:
                #print(f'fail for {fname}')
                logger.info(e)

        raise ScanTypeMismatch(scan_type)


    def _scan_suffix(self, src):
        return src[-6:]
        
    
    def _build_single(self, src, **fmtkeys):
        '''
        Reads a single UDKM experiment dataset.

        Args:
            src: The path / directory where the data is available
              (e.g. "/path/to/data/20250304/174356",
              or "C:\\path\\to\\data\\20250304\\174356").

            **fmtkeys:
        '''
        sfx = self._scan_suffix(src)
        
        for scan_type in ('sRSM', 'tRSS', 'xRSS'):
            try:
                with self._open_as_type(src, scan_type, sfx=sfx) as f:
                    logger.info(f'Trying as type: {scan_type}')
                    fields, images = self._parse_as_type(f, scan_type, base=src)

                    dvars = { k: ('index', v) for k, v in fields.items() }
                    dvars.update({ 'images': (('index', 'x', 'y'), images) })

                    r = xarray.Dataset(data_vars=dvars)

                    if scan_type in ('tRSS',):
                        logger.warning('msg="ACHTUNG, incomplete theta reconstruction"')
                        # ACHTUNG! FIXME: correction needs to be applied here!
                        r['ttheta'] = r.theta * 2

                    return r
                    
            except ScanTypeMismatch as e:
                #logger.warning(e)
                pass

            except Exception as e:
                logger.error(e)
                raise

        raise RuntimeError(f'msg="Don\'t know how to open scan" src={src}')


    def _parse_as_type(self, f, scan_type, base):

        # Parses the indicator lines without leading '% ' (i.e. "Head\tHead\t...\n")
        _make_indicators = lambda line: [
            self.indicators[k.strip()] for k in line.split('\t')
        ]

        # Parses a line like 'Data\tData\tData\t...\n' and assigns
        # the data to the specified indicator list
        _make_data = lambda line, indi: {
            k.strip(): opcast(v.strip())  for k,v in zip(indi, line.split('\t'))
        }

        # Returns image data array from file
        def _make_image(img_path, expected_shape=None):
            try:
                return tifffile.imread(img_path)
            except FileNotFoundError as e:
                missing.append(img_path)
                if expected_shape is None:
                    raise e
                return numpy.full(expected_shape, numpy.nan)

        # Data keys
        lkeys = None

        # Image timing metadata keys
        ikeys = None if scan_type in ('rsm',) \
            else ('start_time', 'end_time', 'image_name')
        imeta = None
        
        img_shape = None
        fields = {}

        missing = []
        images = []

        for line in f:
            if line.startswith('%'):
                # meta-information lines                    
                if lkeys is None:
                    lkeys = _make_indicators(line[2:])
                    continue
                elif ikeys is None:
                    ikeys = _make_indicators(line[2:])
                    continue
                else:
                    imeta = _make_data(line[2:], ikeys)
                    continue
            else:
                # numerical data lines
                pt = _make_data(line, lkeys)
                for k,v in pt.items():                    
                    fields[k] = numpy.append(fields.get(k, numpy.array([])), v)

            # Load the image
            img = _make_image(
                os.path.join(base, f"{imeta['image_name']}"),
                expected_shape=img_shape
            )
            if img is not None:
                img_shape = img.shape
            images.append(img)


        if len(images) > 0:
            logger.warning(f'msg="Some images missing, list follows"')
            if len(missing) > 0:
                msgi = f'{len(missing)}'
                logger.warning(f'msg="Missing images" '
                               f'total="{msgi}" '
                               f'base="{base}"')

        return fields, numpy.concatenate([i[None, ...] for i in images])


ExperimentTemplate = {
    # All the axes -- goniometer first (outer to inner), then detector.
    "goniometerAxes": {
        'theta': 'y-',
    },

    "detectorAxes": {
        'ttheta': 'y-',
    },

    "detectorTARAlign": (0.0, 0.0, 0.0),

    "imageAxes": ("y-", "z+"),
    "imageSize": (195, 487),
    "imageCenter": (97, 241),

    # This could also be used instead of 'imageChannelSize' below.
    # It's the same physical quantity, but in degrees/channel
    # instead of relative length.
    "imageChannelSpan": None,

    "imageDistance": 680.0,
    "imageChannelSize": (0.172, 0.172),  # same unit as imageDistance (mm)

    "sampleFaceUp": 'z-',
    "beamDirection": (1, 0, 0),

    "sampleNormal": (0, 0, -1),

    "beamEnergy": 8048.0,

}


ExperimentTemplate_new = {
    # All the axes -- goniometer first (outer to inner), then detector.
    "goniometerAxes": {
        'theta': 'y-',
    },

    "detectorAxes": {
        'ttheta': 'y-',
    },

    "detectorTARAlign": (0.0, 0.0, 0.0),

    "imageAxes": ("y-", "z+"),
    "imageSize": (195, 487),
    "imageCenter": (97, 241),

    # This could also be used instead of 'imageChannelSize' below.
    # It's the same physical quantity, but in degrees/channel
    # instead of relative length.
    "imageChannelSpan": None,

    "imageDistance": 680.0,
    "imageChannelSize": (0.172, 0.172),  # same unit as imageDistance (mm)

    "sampleFaceUp": 'z+',
    "beamDirection": (1, 0, 0),

    "sampleNormal": (0, 0, +1),

    "beamEnergy": 8048.0,

}

# OBSOLETE -- only for reference purposes
#class DataRun:
#    '''
#    Access to a single UDKM data run.
#
#    Offers easy access to raw data, cooked data, spice data,
#    and experimental information.
#    '''
#    def __init__(self, proposal, key, cache=False):
#        '''
#        Initializes a scan/run management class.
#
#        Args:
#
#          proposal: Parent proposal object
#        
#          key: The scan key
#        
#        '''
#        self._proposal = proposal
#        self._scan_key = key
#
#        try:
#            # If the scan can be interpreted as an integer, we're going to
#            # use it. Otherwise we'll just make up a value. (At this point,
#            # the user should know...)
#            # This functionality is probably better implemented upstream.
#            iscan = int(key)
#        except ValueError:
#            iscan = -1
#
#        #print('new UDKM run at base', self._proposal.scan_url)            
#
#        # proposal.scan_url here is a URL (i.e. with scheme n all), and we
#        # only need the path. At this point, we need to transform everything
#        # into a simple POSIX path.
#        import urllib
#        purl = urllib.parse.urlparse(self._proposal.scan_url.format(scan=key, iscan=iscan))
#        if purl.scheme != "file":
#            raise RuntimeError(f'Invalid scheme "{purl.scheme}" for UDKM legacy')
#        self._base = purl.path
#        
#        self._cache = True
#        self._factory = DatasetFactory()
#        self._raw = None
#        self._cooked = None
#        self._spice = None
#        self._mapper = None
#        self._exp_info = None
#
#    @property
#    def key(self):
#        return self._scan_key
#
#    @property
#    def type(self):
#        return self._factory.typeof(self._base)
#
#    @property
#    def summary(self):
#        return {
#            'type': self.type,
#            'description': self._factory.paramsof(self._base)['Comments']
#        }
#
#    @property
#    def raw(self):
#        if self._raw is None:
#            tmp = self._factory(self._base)
#            if self._cache in (True, 'raw', 'all'):
#                self._raw = tmp
#        else:
#            tmp = self._raw
#        return tmp
#
#
#    def invalidate_raw(self):
#        self._raw = None
#
#
#    @property
#    def exp_info(self):
#        if self._exp_info is None:
#            self._exp_info = ExperimentTemplate.copy()
#        return self._exp_info
#    
#    def cook(self, raw):
#        if self._mapper is None:
#            self._mapper = QMapper(**(self.exp_info))
#
#        run_type = self.type
#        if run_type == 'sRSM':
#            # RSM-data
#            return self._mapper.qmap(self.raw)
#        elif run_type == 'tRSS':
#            # RSS-data
#            c = raw.groupby(raw.delay).map(self._mapper.qmap, dims=('qy', 'qz'))
#            return c
#        else:
#            raise RuntimeError(f'msg="No recipe for run type" type={run_type}')
#
#
#    @property
#    def cooked(self):
#        if self._cooked is None:
#            tmp = self.cook(self.raw)
#            if self._cache in (True, 'all', 'cooked'):
#                self._cooked = tmp
#        else:
#            tmp = self._cooked
#        return self._cooked
#
#
#   @property
#    def spice(self):
#        logger.error('Spice access not available for UDKM')


class LegacyDataProposal(FsDataPropBase):
    def __init__(self, repo, key, url):
        '''
        Access to a UDKM data-pack (i.e. "a day").

        This is for the legacy repository format. Please use `DataProposal`
        and `ProjectRepository instead.

        Args:
          repo: Repository parent object
          key: Identifying key for this proposal
          url: The base URL of the repository
        '''
        super().__init__(repo,
                         key,
                         url,
                         glob='[0-9]'*6,
                         scan_class=DataScan,
                         scan_k2h=lambda p: f'r{p}')


    def _spice_repo(self, url=None):
        # Adding "recipes" and "exp_info", if not found.
        repo = super()._spice_repo(url)
        me = repo.proposal(key=self._proposal_key)
        have = me.collection.view(None)
        if "recipes" not in have.handles():
            from kmc3recipes.cooking import RecipeMap
            me.seed("recipes", __default__=None, **RecipeMap)
        if "exp_info" not in have.handles():
            me.seed("exp_info", **ExperimentTemplate)
        return repo


# Old name
DataPack = LegacyDataProposal


class LegacyRepository(FsDataRepoBase):
    '''
    Access to a complete UDKM data repository in pre-2026 format.

    Expected data layout is "$ENTRY/YYYY/YYYYMMDD/HHmmSS/...".
    The "YYYY/YYYYMMDD" is transtlated to packs (pYYMMDD),
    and the "HHmmSS" part is translated to runs (rHHmmSS).
    '''

    def __init__(self, entry=None):
        '''
        Args:
          entry: entry point for the repository. If not specified,
            the contents of the env-var "UDKM_DATA_REPO" are used.
            Defaults to "." if the env-var is not defined.
        '''
    
        if entry is None:
            entry = os.environ.get('UDKM_DATA_REPO', '.')
        
        logger.info(f'msg="UDKM legacy repo" url="{entry}"')
        
        super().__init__(f'file://{entry}',
                         glob='20??/20??????',
                         proposal_k2h=lambda k: f'p{k[-8:]}')


    def _proposal(self, key):
        # Construct an UDKM proposal based on `DataPack`, from key and URL.
        print(f'New UDKM proposal request {key} at {self.repo_url}')
        return DataPack(self, key, self.repo_url)



class DataScan(DataScanBase):
    '''
    Main scan class for classic UDKM scans.
    '''
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        pu = urllib.parse.urlparse(self.url)
        self._scan_path = pu.path
        self._loader = DatasetFactory()
        
    def _get_raw(self):
        ds = self._loader(self._scan_path)
        po = self._loader.paramsof(self._scan_path)
        # Try this to assign 'po' to data arrays instead of attrs:
        # ds = ds.assign({
        #    'pkey': [k for k in po.keys()],
        #    'params': ('pkey', [v for v in po.values()])
        # })
        return ds.assign_attrs(po)

    def _get_summary(self):
        po = self._loader.paramsof(self._scan_path)
        return {
            'type': self._loader.typeof(self._scan_path),
            'description': po.get('Comments', ''),
            'steps': len(glob.glob(f'{self._scan_path}/*tif')),
        }



class DataProposal(FsDataPropBase):
    def __init__(self, repo, key, url):
        '''
        Access to a UDKM project-style data proposal.
        
        Args:
          repo: Repository parent object
          key: Identifying key for this proposal
          url: The base URL of the repository
        '''
        super().__init__(repo,
                         key,
                         url,
                         glob='20??????/??????',
                         scan_class=DataScan,
                         scan_k2h=lambda p: f'r{p[2:8]}{p[9:]}')


    def _spice_repo(self, url=None):
        # Adding "recipes" and "exp_info", if not found.
        repo = super()._spice_repo(url)
        me = repo.proposal(key=self._proposal_key)
        have = me.collection.view(None)
        #print('checking spice')
        if "recipes" not in have.handles():
            #from kmc3recipes.cooking import RecipeMap
            me.seed("recipes", dict(
                __default__=None,
                tRSM="pymod:///kmc3recipes.cooking#TRSM",
                sRSM="pymod:///kmc3recipes.cooking#sRSM",
                xRSS=None,
                tRSS=None,
                rocking="pymod://kmc3recipes.cooking#sRSM"
            ))
        if "exp_info" not in have.handles():
            me.seed("exp_info", **ExperimentTemplate)
        return repo


class ProjectRepository(FsDataRepoBase):
    '''
    Access to new-style UDKM repository ("project-based").

    The folder structure here is .../YYYYMM_description/ for proposals,
    and .../YYYYMMDD/HHmmSS/ for scans within a proposal.
    '''

    def __init__(self, entry=None):
        '''
        Args:
          entry: entry point (absolute folder!) for the repository. Defaults to
            the contents of the 'UDKM_PROJECT_REPO' environment variable.
        '''

        if entry is None:
            entry = os.environ['UDKM_PROJECT_REPO']
        
        from os import path
        ap = path.abspath(path.expanduser(entry))
        super().__init__(
            url=f'file://{ap}',
            glob='20[0-9][0-9][0-1][0-9]_*',
            proposal_k2h=lambda x: f'{"".join([k.lower() for k in filter(lambda i: i.isupper(), x[7:])])}{x[2:6]}',
            proposal_class=DataProposal,
            proposal_kwargs={}
        )                 

DataRepository = LegacyRepository



#!/usr/bin/python3

## Module to load / save experiment definitions from/to HDF5 Nexus files.

import yaml

def exp_from_yaml(fname):
    '''
    Returns an experiment definition dictionary (as required by QMapper)
    from a YAML file.
    '''

    with open(fname, 'r') as stream:
        try:
            return yaml.safe_load(stream)['experimentSetup']
        except yaml.YAMLError as e:
            logging.error(e)
            raise

def exp2xeuler(exp):
    '''
    "Translates" an experiment setup (or template) into the coordinate
    system naming of the NXeuler Nexus application type. This essentially
    means replacing coordinate letters (x, y, z) with different type
    of coordinate letters.

    The NXxeuler experiment essentially has the following lab system
    definition (see also: https://manual.nexusformat.org/classes/applications/NXxeuler.html#nxxeuler):

     - the Holy Direction of the X-ray Beam is... z. Not x.
    
     - x and y are directions of the detector area / 2D chip.

     - all angles grow counter-clockwise with respect to their
       axis (i.e. they are all "+" direction in xrayutilities lingo).

     - there's no way to specify a direction of the detector pixels,
       i.e. the detector axes are "x" and "y", but pixel 0 apparently
       belongs to the lower value, and pixel N to the higher value
       in that specific direction / on that specific axis.

    Returns an xeuler-compatible experiment template dictionary,
    and a tuple of signs for each of the angles.
    '''
    pass
    # Intermediate name for the axes:
    #  - K: 1st image direction (dimension)
    #  - L: 2nd image direction
    #  - M: X-ray beam direction



from nx5d.repo.base import DataProposalBase
from camagick.source.astor import Processor as Astor
import functools

class DataProposal(DataProposalBase):
    def __init__(self,
                 repo,
                 key,
                 url,
                 scan_class=None,
                 scan_k2h=None):
        '''
        Initializes a filesystem-based proposal walker.

        Args:
          repo: The `DataRepository` object that this proposal
            is part of. This is used because the repository object
            usually also has API elements for access to specific
            services (spice repo, caching etc), which we need to
            pass down.

          key: Identifying information for this proposal inside
            `repo`. Path (inside of `url`) of the group containing
            scans as subgroups.
        
          url: URL (format) of the proposal folder. Use the
            `{scan}` format key to denote how the scan key
            ties into the URL to produce scan URLs. The URL may or
            may not yet have the proposal key filled in. In any case,
            formatting the URL with the proposal key specified in `key`
            (as a replacement for `{proposal}`) will be attempted.

          scan_class: subclass to use for building scan objects
            (typically a subclass of `DataScanBase`, or compatible).

          scan_k2h: key to handle transformator, should match
            `call(key)` and return a python-compatible symbol
            name to use for subscans.
        '''
        super().__init__(repo, key)
        self._k2h = scan_k2h if scan_k2h is not None else lambda x: x
        self._ScanClass = scan_class
        self._astor = Astor(url.format(proposal=key), scans=f'/.keys')


    def _scan_candidates(self):
        self._k2hmap = { str(k):str(self._k2h(k))\
                         for k in self._astor.read()['scans'] }
        return self._k2hmap


    def _scan(self, key):
        return self._ScanClass(self, key)



from nx5d.repo.filesystem import FilesystemWalker, DataRepository
from os import path

def test_walker():
    w = FilesystemWalker('./tests/test_data/udkm/{key}', glob='20??/20??????')
    assert len(w.keys()) > 0

    w = FilesystemWalker('./tests/test_data/spice/{key}')
    print(w.keys())

    
def test_repo1():
    r = DataRepository(url=path.abspath('./tests/test_data/udkm/{proposal}/{scan}'),
                       glob='20??/20??????',
                       proposal_k2h=lambda x: f'p{x[7:]}')
    print('proposals in repo:', r.all())
    assert len(r.all()) > 1
    assert '2025/20250408' in r.all()

    
def test_scan1():
    pass



#!/usr/bin/python3

import pytest
import nx5d.h5like.fio as p3

@pytest.fixture
def fio_file():
    return 'tests/test_data/fioh5/m3_2507_00744.fio'


@pytest.fixture
def fio_tiff():
    return 'tests/test_data/fiotiff/s1_1161_00484.fio'

def test_fio_parser(fio_file):
    ''' Reading/parsing a FIO file '''
    
    pars, data = p3.read_fio(fio_file)

    assert data['om'].shape[0] == 61
    assert type(pars['om']) == float
    
    #print("Got FIO:", data)

    
def test_fio_loadh5(fio_file):

    fh = p3.FioH5(fio_file)

    assert "fio" in fh.keys()
    
    assert len(fh['fio'].keys()) == 2
    assert len(fh['fio/parameters']) > 10
    assert "om" in fh['fio/parameters'].keys()
    assert "om" in fh['fio/data'].keys()

    assert "lambda_00000" in fh.keys()
    
    a = fh['lambda_00000']['entry']
    b = fh['lambda_00000/entry']

    assert a.keys() == b.keys()

    data = fh['lambda_00000/entry/instrument/detector/mock_data'][()]

    # Very simple test of retrieving the data -- we know this to be the shape
    # of 'mock_data' of the test dataset. Real datasets have a different field
    # ("data") with different shapes, of course.
    assert data.shape == (61, 100, 100)

    assert data.shape[0] == fh['fio/data/om'].shape[0]


def test_fio_loadtiff(fio_tiff):
    
    f5 = p3.FioTiff(fio_tiff, 'fio')

    # This is where we store the TIFF dataset
    assert "fio" in f5
    assert "p100k" in f5['fio']
    assert len(f5["fio/p100k/data"].shape) == 3

    # This is real data, sum is somewhere around 25k or so.
    assert f5["fio/p100k/data"][0:2,...].sum() > 0



#!/usr/bin/python3

from scipy import ndimage
import numpy as np
from xarray import DataArray
import logging

logger = logging.getLogger(__name__)

__all__ = [ "series_cntrmass", "sqdistance",  "stdwidth1d", "series_stdwidth1d" ]

'''
Helpers for X-ray diffraction data analysis
'''

def series_cntrmass(data, series=None):
    ''' Calculates N-1 dimensional center of mass for an N-dimensional array.

    Args:
    
        data: This is expected to be an `xarray` with at least 2
           named dimensions, or a list of N-dimensional arrays.

        series: If `data` is an `xarray`, this is expected to be a string
           naming the axis along which to break data apart and form
           a series of.
           
           If data is a `numpy` array or a regular (nested) list,
           `series` is expected to be an axis index.
           
           If data is an iterable (e.g. a list of arrays), then `series`
           must remain `None`, and the `data` iterable already defines
           the series.

           In all cases, if `series=None` (the default), the center of
           mass series will be built along the 1st dimension.
    
    Returns: a 2D `xarray.DataArray`. The first dimension has the same length as the
        length of the series (i.e. the length of the `series` dimension).
        The 2nd dimension hs a length of N-1, where N is the number of dimensions
        of the original data series.
        Each entry being an N-1 dimensional coordinate of the center of mass in 
        the respective plane.
    '''
    
    
    # The difficult part is going to be that ndimage's center_of_mass()
    # returns a fractional (!) positional (!!) index, i.e. one that's
    # related to the position.
    
    # Push the dimension we want to keep to the begging (that's
    # apparently the only one we can iterate over)
    if series is not None:
        view = data.transpose(series, ...)
    else:
        view = data
        
    index_list = []

    if hasattr(view, 'dims'):
        # ...if input is a single DataArray
        
        if len(view.shape) <= 1:
            raise RuntimeError(f'Input data must be a sequence of datasets (i.e. multiple arrays),'
                               f' or a multi-dimensional array.')
        
        other_axes = [o for o in view.dims[1:]]
    else:
        # ...if input series is actually a list of (1D) datasets
        other_axes = list(next(iter(data)).dims)

    
    ## iterate along `axis` and collect all the COMs in index_list
    for img in view:
        try:
            com = ndimage.measurements.center_of_mass(img.data)
            ci  = np.floor(com).astype(int)
            cf  = com-ci
        except ValueError as e:
            # All-zero datasets produce errors (there's no useful definition
            # of a COM on an all-zero dataset). It's the upper layer's responsibility
            # to fix this, but we want to give the user a little bit more information
            # to go on.
            logging.error(f'Error: {str(e)}: this is probably because you have all-zero datasets')
            raise
        
        # iterate over each COM coordinate of the N-1 dim 
        #index = [ view.coords[c].values[i] + f*(img.coords[c].values[i+1]-img.coords[c].values[i])
        #             for c,i,f in zip(other_axes, ci, cf) ]

        index = [ img.coords[c].values[i] + f*(img.coords[c].values[i+1]-img.coords[c].values[i])
                     for c,i,f in zip(other_axes, ci, cf) ]
        
        index_list.append(index)

    # We return a 2D DataArray. In the first dimension, we have one entry
    # for each of the COM sets. We're naming this as the original axis
    # was named (if 'series' is defined), or simply 'plane' if the data sets
    # were passed on as a list.
    # The 2nd dimension is a string list of all the other dimension names
    # (so that we know how to attribute the COM components).
    
    if series is not None:
        series_coords = view.coords[series].values
        series_dim = series
    else:
        series_coords = [i for i in range(len(view))]
        series_dim = 'plane'
        
    coords = {
        series_dim: series_coords,
        'axis': other_axes
    }
        
    return DataArray(data=index_list, coords=coords)


def sqdistance(*axes, center, shape=None):
    '''
    Takes a list of axes and a location desginated `center` and returns
    a N-dimensional array with all the squared distances from `center`
    to each of the points. The array dimensions are assumed to be
    equivalent in size.
    
    Parameters:
    
      - `*axes`: Axis objects, one per parameter. One axis object
        can either be an `xarray` coordinates tuple `(dim, values)`
        with `dim` as a string identifyer and `values` as the axis
        values; or they can be plain iterables (regular `numpy` 1D
        arrays or lists).
        
      - `center`: needs to be 1-dimensional vector with N values,
        if `axes` items are regular `numpy` arrays or iterables.
        Otherwise 
        It represents the point relative to which the squared distance
        of all other points is calculated.
        Despite its name it does not need to be, and typically is
        not, the actual center of the field.
        
      - `shape`: Instead of explicitly specifying a list of axes,
        a single vector continaing an N-dimensional array shape
        can also be specified. In that case, the axes will be
        constructed as regular indices ranging from `0` to `len(shape[n])`
        
    Returns: N-dimensional array of size `len(axes[0]) * len(axes[1]) * ... * len(axes[n])`
    contraining the squared distance to `center` for each point.
    '''
        
    if (not axes or len(axes) == 0) and shape is not None:
        # numpy array style shape. We don't actually have axis
        # values, we need to build them first. `center` is a
        # 1D-array (one coordinate per axis) we can juse for
        # indexing.
        axvalues = [[i for i in range(s)] for s in shape] 
        cvalues = tuple(center)
        return_numpy = True
        
    elif isinstance(axes[0], tuple) and len(axes[0]) == 2 and isinstance(axes[0][0], str):
        # Named axes, `xarray` or `pandas` style. Each axis element
        # is either one `xarray` Axis object (with values in the .values
        # property) or a (name, values) axis tuple.
        # `center` is either a dict, or a list of (key, val) tuples.
        axvalues =  [ (x[1].values if hasattr(x[1], "values") else x[1])   for x in axes]
        cobj = dict(center)
        cvalues = tuple([cobj[x[0]] for x in axes])
        return_numpy = False
        
    else:
        # numpy array, explicit axes. Each axis is a 1D-iterable
        # and `center` is a 1D-tuple (one coordinate per axis).
        axvalues = axes
        cvalues = center
        return_numpy = True
    
    sqdist = np.abs(axvalues[0] - cvalues[0])**2
    for ax,c in zip(axvalues[1:], cvalues[1:]):
        sqdist = sqdist[...,None]
        sqdist = sqdist + np.ones(sqdist.shape) * (ax-c)**2
    
    if return_numpy:
        return sqdist
    
    return DataArray(data=sqdist, coords=[(x[0], x[1].values) for x in axes])
    
    
def stdwidth1d(data, center):
    ''' Calculates the standard deviation a.k.a. "peak width".

    This is defined as the square-root of the
    [variance](https://en.wikipedia.org/wiki/Variance#Discrete_random_variable)
    of an experimental data array with respect to the value located in the
    same array at index position called `center`.

    For 1-dimensional arrays of discrete points, the variance is essentially
    defined as `(Xi-M)**2`, i.e. the squared sum of the differences of the
    poins `Xi` from a mean value `M`. It's a measure of how much each
    value differs from the mean.

    For N-dimensional data, the statistical reasoning is more complex and
    evolves around [the covariance matrix](https://en.wikipedia.org/wiki/Covariance_matrix).
    This function does *not* implement the covariance matrix version
    (serach for `stdvariance()` instead), but only the simple 1D version.
    If fed with a multiple N-dimensional input, it *still* only calculates
    one variance, which can essentially be interpreted as an "average variance"
    along all dimensions.
    
    Args:
    
        data: The data for which to calculate the standard width.
          This can either be:
            - an N-dimensional `numpy` array
            - an N-dimensinonal `xarray.DataArray`
        
        center: The center of mass relative to which to calculate the
          deviation. This can be:
    
            - an N-tuple capable of indexing `data`. This is expected,
              if `data` is an numpy array.

            - a dictionary with dimension names as keys, and positions
              (coordinate lists/arrays) as values, or...

            - ...a single `(dim, value)` tuple, or...

            - ...an `xarray.DataArray` with positions / center-of-mass
              coordinates.
      
    Returns: One single value, represending the sum of suqare distances
    weighted by the function value at the respective place, if
    `axis=None`. If `axis` is specified, then the result will be a
    1-dimensional array containing standard widths along `axis`.
    '''
    
    # Preparing the coordinate axis array of `data' -- for an xarray
    # we just build (name, values) tuples of its dims; for numy arrays,
    # we generate an integer index corresponding to the length.
    if hasattr(data, "coords"):
        axes = [(d, data.coords[d]) for d in data.dims]
    else:
        axes = [range(i) for i in data.shape]
        
    pos = center
        
    sqdist = sqdistance(*axes, center=pos)
    sqvar  = (sqdist * data / data.sum()).sum()
    
    return np.sqrt(sqvar)


def series_stdwidth1d(data, center, axis=0):
    '''
    Wrapper for `stdwidth1()` to act on a collection of data sets at once.
    #
    '''
    pass



import pytest, json, glob, os, pprint
from nx5d.spice import  spice_schema, FsSpiceRepository, SpiceCollection, \
    SpiceView, SpiceRevisionConflict, BranchView, MemorySpiceRepository, \
    SpiceProposal

def test_load():
    base = os.path.abspath('./tests/test_data/spice/pack1')

    col = SpiceCollection()
    
    for p in glob.glob(f'{base}/*'):
        with open(p, 'r') as f:
            data = json.load(f)
            col.add(data)

    vs = col.view("r01", 'straight')
    assert hasattr(vs, "straight")
    assert hasattr(vs.straight, "param")
    assert vs.straight.param == 3.14

    with pytest.raises(SpiceRevisionConflict):
        vc = col.view("r01", 'conflicted')

    v = col.view("r01", ignore_conflict=True)
    assert hasattr(v, 'straight')
    assert hasattr(v, 'conflicted')

    assert (not v.straight.conflicted())
    assert (v.conflicted.conflicted())

    print (v.straight.meta())
            
    v2 = col.view("r05", 'straight')
    v3 = col.view("r10", 'straight')

    assert v2.straight.param == vs.straight.param
    assert v3.straight.param != vs.straight.param

    print(v2.straight)
    print(v3.straight)
    
    # Test retrieving the seed
    s1 = col.view(None, 'straight')
    print('anchor:', s1)

    
def test_fs_spice():
    sr = FsSpiceRepository(os.path.abspath('./tests/test_data/spice/{proposal}'))
    scan1 = sr.pack1.collection.view("r01", ignore_conflict=True)

    assert "straight" in sr.pack1.collection.handles()
    assert hasattr(scan1, "straight")


def test_find():
    sr = FsSpiceRepository(os.path.abspath('./tests/test_data/spice/{proposal}'))
    
    f = sr.pack1.collection.find(startswith=True, uuid='f965')
    assert len(f) == 1

    f2 = sr.pack1.collection.handles('straight')
    pprint.pprint(f2)


def test_memrepo_empty():
    repo = MemorySpiceRepository()
    assert len(repo.all()) == 0

    repo.new('myprop')
    assert 'myprop' in repo.all()
    
    p = repo.myprop    
    assert isinstance(repo.myprop, SpiceProposal)


@pytest.fixture
def memrepo():
    t = FsSpiceRepository(os.path.abspath('./tests/test_data/spice/{proposal}'))
    return MemorySpiceRepository(t)

    
def test_memrepo_template(memrepo):
    assert "pack1" in memrepo.all()
    #assert "pack2" in memrepo.all() ## missing in repo?


def test_seed(memrepo):
    
    # Should succeed (new type)
    memrepo.pack1.seed('straight2', data=3.14, idem=False)
    v = memrepo.pack1.collection.view('r01', 'straight2')
    assert hasattr(v, 'straight2')
    assert 'data' in v.straight2.data()

    # Should fail ('straight' already exists as a spice type)
    with pytest.raises(ValueError):
        memrepo.pack1.seed('straight', data=3.14, idem=False)

    # Should work (idem parameter)
    exists = memrepo.pack1.seed('straight', data=3.14, idem=True)
    assert exists['uuid'] == memrepo.pack1.collection.view(None, 'straight').straight.meta('uuid')
    assert exists['uuid'] == memrepo.pack1.collection.view('r01', 'straight').straight.meta('uuid')


def test_anchor(memrepo):
    v1 = memrepo.pack1.collection.view(None, 'straight').straight.obj()
    v2 = memrepo.pack1.anchor('r02', 'straight')
    
    print('v1', v1)
    print('v2', v2)
    assert v1['data'] == v2['data']

    v3 = memrepo.pack1.anchor('r02', 'straight')
    assert v2['data'] == v2['data']
    assert v2['uuid'] == v2['uuid']


def test_update(memrepo):
    v1 = memrepo.pack1.update('r02', 'straight', param=1)
    
    print('update:', v1)
    assert v1['data']['param'] == 1
    
    v2 = memrepo.pack1.update('r02', 'straight', param=1)
    print('idem update:', v2)
    assert v2['uuid'] == v1['uuid']

    v3 = memrepo.pack1.update('r02', 'straight', param=2)
    print('overwrite:', v3)
    assert v2['uuid'] != v3['uuid']
    assert v3['data']['param'] == 2
    assert v2['data']['param'] != v3['data']['param']

    v4 = memrepo.pack1.update(v3['uuid'], param=3)
    print('update by uuid:', v4)
    assert v4['uuid'] != v3['uuid']
    assert v4['data']['param'] != v3['data']['param']
    assert v4['data']['param'] == 3



from nx5d.repo.base import DataRepoBase, DataProposalBase
import parse, os, logging

import glob as glb

from functools import partial

logger = logging.getLogger(__name__)

# Implements filesystem-based proposal and data reading.

class FilesystemWalker:
    '''
    Class to conveniently iterate through filesystem folders.

    The asumption is that data is stored in a filesytem directory
    structure.

    Args:
      fmt: starting point, containing a format key as the
        place where subentries should be expected.
    '''
    def __init__(self, fmt, glob=None):

        if glob is None:
            glob = '*'

        self._format = fmt
        self._glob = fmt.format(key=glob)


    def _path2key(self, p):
        '''
        Returns the key part from a path.
        '''
        r = parse.parse(self._format, p)
        if (r.named is None) or ('key' not in r.named):
            raise RuntimeError(f'msg="Cannot parse key from path" '
                               f'path="{p}" '
                               f'format="{self._format}"')
        return r.named['key']


    def _key2path(self, key):
        '''
        Given a sub-group key, returns the corresponding path.
        '''
        return self._format.format(key)


    def _list_entries(self):
        # Lists all entries below `base`.
        # Override this if the backend is not an actual filesystem.
        return glb.glob(os.path.join(self._glob))


    def keys(self):
        return [
            self._path2key(i) for i in self._list_entries()
        ]


    def entry(self, name):
        path = self._key2path(name)
        return self._subwalker(os.path.join(self._entry, path))

    
class GenericDataWalker:
    # Base class for filesystem-globbing listing for various resources.
    #
    # This is used as a base class for various filesystem-walking
    # containers, like DataRepository and DataProposal (here),
    # but also `nx5d.spice.FsSpiceRepository`.
    #
    # This base implements mechanisms for URL loading and listing
    # (file/s3/https/tield come to mind), and transforming of path
    # keys <-> handles, as well as delegating __getattr__() calls
    # to sub-classes.
    #
    # Note that the subclasses `DataRepository`, respectively
    # `DataProposal` don't necessarily need to be used together. In
    # fact, most cases will likely have a FS-based `DataRepository`,
    # while having something else entirely (e.g. an ArrayAccess-based
    # `DataProposal`).
    def __init__(self,
                 url,
                 glob=None,
                 k2h=None,
                 h2k=None,
                 subnode_class=None
                 ):
        '''
        Initializes access to a filesytem-based repository reader.

        The general idea is that the repository location is
        a filesystem or filesystem-like entity able to deliver
        all proposals by a globbing pattern, and/or to retrieve
        a _specific_ proposal by filling in a formatting key.

        Args:
        
          url: string format for a full, typical item key path
            (e.g. "file:///mnt/data/proposals/{key}/")
        
          glob: globbing pattern for the data item, if different
            from `*`.

          subnode_class: class to initialize the return value
            of `._subnode()` with. The sole argument of the
            proposal class is the key, e.g. "proposal url"
            or "scan url".

          k2h: if not `None`, expected to be a callable `call(key)`
            which will transform an on-disk proposal key into a
            Python compatible handle.

          h2k: if not `None`, expected to be a callable `call(handle)`
            that will transform a Python handle into an on-disk
            proposal key.
        '''

        self._k2h = k2h if k2h is not None else lambda x: x
        self._h2k = h2k

        # for later use when creating subnodes
        self._base_url_format = url

        import urllib
        purl = urllib.parse.urlparse(url)
        
        self._walker = self._init_walker(purl.scheme,
                                         fmt=purl.path,
                                         glob=glob)
        self._SubnodeCls = subnode_class


    def _init_walker(self, scheme, **kw):
        return {
            'file': FilesystemWalker
        }[scheme](**kw)

    def _key2handle(self, k):
        if self._k2h is None:
            raise RuntimeError('no key->handle transform available')
        return self._k2h(k)

    def _handle2key(self, h):
        if self._h2k is None:
            return self._hkmap[h]
        return self._h2k(h)

    def _subkeys(self):
        self._khmap = {
            k:self._key2handle(k) for k in self._walker.keys()            
        }
        self._hkmap = {
            h:k for k,h in self._khmap.items()
        }
        return self._khmap.keys()

    
    def key_handle_map(self):
        self._subkeys() # regenerate map
        return self._khmap


    def _subnode(self, key):
        sub_url = self._base_url_format.format(key=key)
        #print("Walker Subnode Class:", self._SubnodeCls)
        return self._SubnodeCls(sub_url)


class DataRepository(GenericDataWalker, DataRepoBase):
    def __init__(self,
                 url,
                 glob=None,
                 proposal_k2h=None,
                 proposal_class=None,
                 proposal_kwargs=None,
                 **repo_kwargs):
        '''
        Initializes an on-disk proposal manager.

        Args:
          url: Path format for finding the proposals. Must contain
            `{proposal}` as a formatting key.
          glob: globbing pattern for the proposal keys,
            defaults to `*` if not specified.
          proposal_k2h: key-to-handle translation, defaults
            to a `lambda x: x`. Should otherwise be a callable,
            which, wen called with the proposal key as a parameter,
            returns a uniqe handle.
          proposal_class: class to instantiate on every new .proposal()
            request. If set to "auto" or "same", this
            specific repository implementation will just assume that
            the proposal subclass is a filesystem walker based
            class. You can pass a different kind of class, but it
            must have the same initialization parameters as
            the `FsDataProposal` class. Alternatively, ignore
            this or set to `None` (the default), and overwrite
            the `._proposal()` member function.
          proposal_kwargs: additional keyword arguments to pass on to
            the initialization of the proposal class.
          **repo_kwargs: additional keyword arguments will be
            passed to the `DataRepoBase` subclass.
        '''
        DataRepoBase.__init__(self, url, **repo_kwargs)

        import urllib
        purl = urllib.parse.urlparse(url)
        self._proposal_minimized = self._minimize_path(purl.path)

        # This is the smaller proposal-only URL. Since this is a
        # filesystem proposal lister, we ignore everything that isn't
        # host or path (intra-document stuff is not handled by this
        # DataProposal lister)
        scheme = purl.scheme if (purl.scheme is not None) and len(purl.scheme)>0 \
            else 'file'
        self._proposal_min_url = \
            f'{scheme}://'\
            f'{purl.netloc}/'\
            f'{self._proposal_minimized}'
        
        # This is what we'll pass on to the filesystem walker,
        # repo key here is "...{key}..."

        if self._proposal_minimized != purl.path:
            logger.info(f'msg="Minimized proposal URL" '
                        f'path={self._proposal_min_url}')
        
        self._walker_base_format = self._proposal_min_url.format(
            proposal="{key}"
        )
        
        GenericDataWalker.__init__(
            self,
            self._walker_base_format,
            glob=glob,
            k2h=proposal_k2h,
            h2k=None, # Reverse lookup handle->key done by superclass
            subnode_class=None, # Disable sub-walking.
                                # Need to overwrite ._proposal()!
        )

        self._ProposalSubclass = proposal_class
        self._proposal_kwargs = proposal_kwargs \
            if proposal_kwargs is not None else {}


    def _minimize_path(self, original_path, **keywords):
        # Calculates the minized {proposal}-only path,
        # by substituing increasingly larger path elements,
        # until at {proposal} gets substituted.
        from os import path
        head = original_path
        while len(head) > 0:
            try:
                head.format(proposal='jinkies!')
                return head
            except KeyError: pass
            head, tail = path.split(head)
        raise RuntimeError(f'unusable proposal path: {original_path}')
        

    @property
    def proposal_url(self):
        return self._proposal_min_url

    def all(self):
        return self.key_handle_map()

    def keys(self):
        return self.key_handle_map().keys()

    def handles(self):
        return self.key_handle_map().values()

    def _proposal(self, key):
        if self._ProposalSubclass == 'auto':
            self._ProposalSubclass = DataProposal

        if self._ProposalSubclass is None:
            raise RuntimeError(f'No proposal subclass available')
        from collections import defaultdict
        dd = defaultdict(lambda x: x)
        subpath = self._repo_url.format(
            proposal=key, scan='{scan}'
        )
        return self._ProposalSubclass(repo=self, key=key, url=subpath,
                                      **self._proposal_kwargs)


class DataProposal(GenericDataWalker, DataProposalBase):
    def __init__(self,
                 repo,
                 key,
                 url,
                 glob=None,
                 scheme=None,
                 scan_class=None,
                 scan_k2h=None):
        '''
        Initializes a filesystem-based proposal walker.

        Args:

          repo: Repository object (parent)

          key: Identifying information for this proposal
        
          url: URL (format) of the proposal folder. Use the
            `{scan}` format key to denote how the scan key
            ties into the URL to produce scan URLs. The URL
            may or may not have the specific proposal key
            already baked into the URL. In any case, replacing
            the format key `{proposal}` for the specified `key`
            parameter will be attempted.
        
          glob: globbing/search pattern for the scans, if
            different from "*".

          scheme: this will generally be obtained from the URL,
            or will default to "file". Specify something else
            ("s3", "https", "tiled"...?) manually if you need.

          scan_class: subclass to use for building scan objects
            (typically a subclass of `DataScanBase`, or compatible).

          scan_k2h: key to handle transformator, should match
            `call(key)` and return a python-compatible symbol
            name to use for subscans.
        '''
        #print(f"New filesystem proposal: key={key} url={url}")
        DataProposalBase.__init__(self, repo, key)

        # This is the URL format for Run sub-nodes; need to fill in the
        # {proposal} key, but keep open all other keys (in particular
        # {scan} or {iscan}).
        from collections import defaultdict
        class _dd(defaultdict):
            def __missing__(self, key):
                return f'{{{key}}}'
        self._scan_url = url.format_map(_dd(proposal=key))

        walker_url_base = url.format(proposal=key, scan="{key}")

        #print(f'walker base: {walker_url_base}')

        self._ScanClass = scan_class
        GenericDataWalker.__init__(self,
                                   walker_url_base,
                                   glob=glob,
                                   k2h=scan_k2h)


    @property
    def scan_url(self):
        '''
        Returns a URL pattern that contains `{scan}` or where the scan
        info should go.
        '''
        return self._scan_url


    def all(self):
        return self.key_handle_map()

    def keys(self):
        return self.all().keys()

    def handles(self):
        return self.all().values()

    def _scan(self, key):
        return self._ScanClass(self, key)




#!/usr/bin/python3

from nx5d.xrd.xrd_helpers import *
import pytest

import numpy as np
from xarray import DataArray

def test_stdwidth1d_ndarray(test_ndarray):
    '''
    Test for stdwidth()
    '''

    center = (np.array(test_ndarray.shape)/2).astype(int)
    
    v = stdwidth1d(test_ndarray, center)

    # We're measuring the variance ("width") with regards to the
    # actual center of the data patch, so this should always essentially
    # be smaller than half the field.
    # This being random data, it *should* have a width larger than 1
    # grid point, though...
    assert (v <= np.abs(center.max()))
    assert (v >= 1.0)

    
def test_stdwidth1d_xarray(test_xarray):
    '''
    Test for stdwidth()
    '''

    # alias
    dta = test_xarray

    center = [ (x, dta.coords[x].values[0]+(dta.coords[x].values[-1]-dta.coords[x].values[0])/2 ) \
                     for x in test_xarray.dims ]
    
    v = stdwidth1d(test_xarray, center=dict(center))

    # We're measuring the variance ("width") with regards to the
    # actual center of the data patch, so this should always essentially
    # be smaller than the field.
    c_coord = np.array([dta.coords[x].values[-1]-dta.coords[x].values[0] for x in dta.dims])
    assert (v <= np.abs(c_coord.max()))


    
@pytest.fixture
def test_ndarray():
    ''' Returns a an ndarray of variable size and shape (i.e. dimensions) '''

    ## ...up to 7 dimensions should prove the point.
    ## Configurable for debugging.
    
    min_dims = 2
    max_dims = 2 #7

    min_size = 2
    max_size = 10 #20
    
    dims = min_dims + int(np.random.random() * (max_dims-min_dims))
    shp  = min_size + (np.random.random(dims) * (max_size-min_size)).astype(int)
    return np.array(np.random.random(shp.prod())).reshape(shp)


@pytest.fixture
def test_ndcenter(test_ndarray):
    ''' Returns a random 'center' coordinate that lies within the array '''
    return (np.random.random(len(test_ndarray.shape)) * test_ndarray.shape).astype(int)
    

def test_sqdist_ndarray(test_ndarray, test_ndcenter):
    '''
    Tests sqdistance() on an ndarray (i.e. no named axes)
    '''
    
    assert ((test_ndcenter < test_ndarray.shape).all())
    
    sqd1 = sqdistance(center=test_ndcenter, shape=test_ndarray.shape)
    
    axlist = tuple([range(s) for s in test_ndarray.shape])
    sqd2 = sqdistance(*axlist, center=test_ndcenter)

    # No distance can be longer than the diagonal across all dimensions
    assert (sqd1.max() <= (np.array(test_ndarray.shape)**2).sum())
    assert (sqd2.max() <= (np.array(test_ndarray.shape)**2).sum())
    
    # The square distance result has exactly one point for every input point
    assert (sqd1.shape == test_ndarray.shape)
    assert (sqd2.shape == test_ndarray.shape)
    
    assert ((sqd1[tuple(test_ndcenter)] == sqd2[tuple(test_ndcenter)]).all())
    
    # Distance at center is always zero
    assert((sqd1[tuple(test_ndcenter)] == 0).all())
    assert((sqd2[tuple(test_ndcenter)] == 0).all())
                                                                      
                                                                      
def test_sqdist_ndtuple(test_ndarray, test_ndcenter):
    '''
    Test whether sqtistance() accepts tuples as arguments
    '''
    axlist = tuple([range(s) for s in test_ndarray.shape])
    sqd = sqdistance(*axlist, center=tuple(test_ndcenter))
    assert (sqd.shape == test_ndarray.shape)
                                                                      
                                                                      
@pytest.fixture
def test_xarray(test_ndarray):
    ''' Creates an `xarray` with named axes based on a random ndarray '''
    
    axis_names = ['physics', 'trustee', 'warning', 'ethnic',
                  'tree', 'text', 'contraction', 'shell' ]
    
    return DataArray(data=test_ndarray,
                     coords=[np.array(range(s))*0.1-(s/20) for s in test_ndarray.shape],
                     dims=axis_names[:len(test_ndarray.shape)])
    

@pytest.fixture
def test_xcenter(test_xarray):
    '''
    Generates a random single-point center coordinate based on an
    array with named axes. The coordinate tuple has named dimensions.
    '''
    
    pos = {}
    for d in test_xarray.dims:
        ax = test_xarray.coords[d].values
        r = np.random.random()
        c = ax[0] + r*(ax[-1]-ax[0])
        pos[d] = float(c)

    return pos


def test_sqdist_xarray(test_xarray, test_xcenter):
    '''
    Tests the sqdistance() function with xarray-like named axes.
    '''
        
    ## This is for more specific testing / debugging.
    #shape  = (5, 5)
    #data   = np.ones(shape)
    #axes   = [(np.array(range(s))-s/2)*0.1 for s in shape]
    #center = {'x': 0, 'y': 0 }
    #xdata  = DataArray (data=data, coords=axes, dims=['x', 'y'])
    
    center = test_xcenter
    xdata  = test_xarray
    
    sqd = sqdistance(*xdata.coords.items(), center=center)
    
    assert (sqd.max() <= (np.array(xdata.shape)**2).sum())
    
    # The square distance result has exactly one point for every input point
    assert (sqd.shape == xdata.shape)
    
    # Calculate distance (N-dim diagonal) between two ajacent axis coordinates.
    # This will give us a very crude estimate of an error measure which is
    # (1) smaller than a pixel, but
    # (2) still large enough such that an interpolation of the array value
    # at that specfic point will still fit well within that error value.
    
    pix_diag = np.sqrt(np.array([ (x[1].values[1]-x[1].values[0])**2 for x in xdata.coords.items() ]).sum())
    maxerr = pix_diag*0.5 / len(xdata.shape)
    
    #print ("### distances:\n", np.sqrt(sqd.values))
    #print ('### pixdiag:', pix_diag)
    #print ('### maxerr:', maxerr)
    #print ('### value(s) at center:', sqd.interp(center).values)
    #print ('### center coordinates:', center)
        
    # Distance at center is always zero
    assert( (sqd.interp(center) < maxerr).all())

    # Make sure sqdistance() accepts a list of tuples as center coordinates
    sqd2 = sqdistance(*xdata.coords.items(), center=center.items())
    assert ((sqd2 == sqd).all())

    # Make sure sqdistance() accepts a dict() center coordinates
    sqd3 = sqdistance(*xdata.coords.items(), center=dict(center.items()))
    assert ((sqd3 == sqd).all())

    # Make sure sqdistance() accepts a list of tuples as axes
    #sqd3 = sqdistance(*tuple( [(d,xdata.coords[d]) for d in xdata.dims] ),
    #                  center=center)
    #assert ((sqd2 == sqd3).all())



import pytest, json, glob, os, pprint, uuid
from nx5d.slim import Application
from nx5d.slim import _json_from_args

def test_jargs():
    
    j = _json_from_args('foo=42',
                      'bar="bar"',
                      'obj={"key": "value", "array": [1, 2, 3]}')
    assert j['foo'] == 42
    assert j['bar'] == "bar"
    assert j['obj']['key'] == "value"
    assert j['obj']['array'] == [1, 2, 3]

    j = _json_from_args('./tests/test_data/json/offsets.json')
    for i in ('chi', 'phi', 'theta', 'tth'):
        assert i in j
    assert j['chi'] == 0

    j = _json_from_args('./tests/test_data/json/offsets.json', 'chi=1.0')
    for i in ('chi', 'phi', 'theta', 'tth'):
        assert i in j
    assert j['chi'] == 1.0

    j = _json_from_args('chi=2.0', './tests/test_data/json/offsets.json')
    for i in ('chi', 'phi', 'theta', 'tth'):
        assert i in j
    assert j['chi'] == 2.0

    j = _json_from_args('chi=2.0',
                        './tests/test_data/json/offsets.json',
                        './tests/test_data/json/energy.json')
    for i in ('chi', 'phi', 'theta', 'tth', 'energy'):
        assert i in j
    assert j['chi'] == 2.0

    with pytest.raises(RuntimeError):
        j = _json_from_args('nonexistent')

    pprint.pprint(j)


@pytest.fixture
def env():
    import os
    return {
        "NX5D_SPICE_REPO": os.path.abspath("./tests/test_data/spice/{proposal}")
    }

def test_proposals(env):
    # Lists proposals, no output    
    o = Application(args=["slim", "proposals?"], env=env)._run()
    print('result:')
    print(o)
    assert o == None


def test_list(env):
    o = Application(env=env, args=[
        "slim", "pack1", "list"
    ])._run()
    print('result:')
    print(o)
    assert len(o) == 2 # 2 anchor points for 'pack1'


def test_view(env):
    o = Application(env=env, args=[
        "slim", "pack1", "view", "r01"
    ])._run()
    print('result:')
    print(o)
    assert len(o) == 3 # 3 spice types for 'pack1'


def test_update(env):
    o = Application(env=env, args=[
        "slim", "pack1", "update", "straight@r01", "--dry"
    ])._run()
    print('result:')
    print(o)
    u = uuid.UUID(o)
    

def test_seed(env):

    # Should succeed
    o = Application(env=env, args=[
            "slim", "pack1", "seed", "straight2", "data=3.14", "--dry"
        ])._run()
    print('result:')
    print(o)
    u = uuid.UUID(o)

    # Should fail ('straight' already exists)
    o = Application(env=env,
        args=[
            "slim", "pack1", "seed", "straight", "data=3.14", "--dry",
        ])._run()
    print('result:')
    print(o)
    with pytest.raises((ValueError, AttributeError)):
        uuid.UUID(o)
    

def test_anchor(env):
    o = Application(env=env,
        args=[
            "slim", "pack1", "anchor", "r02", "straight", "--dry"
        ])._run()
    print('result:')
    print(o)
    u = uuid.UUID(o)
    




import logging

from itertools import takewhile
from more_itertools import before_and_after

logger = logging.getLogger(__name__)

class UnknownScanType(Exception): pass

class DataRepositoryBase:
    '''
    Abstract base class for `DataRepository` type of objects.

    The main purpose of a `DataRepository` is to provide listing
    and basic access to proposal data via `.all()`,
    respectively `.proposal()`.

    We'll also reuse the `.__dir__()` Python-builtin for
    listing the proposals, and the `.__getattr__()` for
    returning a proposal object.
    '''

    def __init__(self, url, spice_url=None):
        self._repo_url = url

        # We keep a local cache of proposal objects that we create.
        # This is because associated with a proposal, there may be
        # other associated (e.g. spice), and the expectation is that
        # <repo>.<proposal> conserves state.
        self._proposal_cache = {}

        self._spice_url = spice_url

    @property
    def spice_url(self):
        return self._spice_url

    @property
    def repo_url(self):
        '''
        The full repository URL as specified by user
        '''
        return self._repo_url


    def all(self):
        '''
        Returns a key->handle map for all
        proposals managed by this repository.
        '''
        raise RuntimeError('not implemented')


    def keys(self):
        return self.all().keys()


    def handles(self):
        return self.all().values()


    def _proposal(self, key):
        '''
        Returns a `DataProposal` object for the proposal of this
        repo designated by `handle` or by `key`.
        '''
        raise RuntimeError('not implemented')


    def proposal(self, *, handle=None, key=None):
        if key is None:
            try:
                key = {h:k for k,h in self.all().items()}[handle]
            except KeyError as e:
                logger.warning(f'msg="Proposal not present in spice repo" '
                               f'handle="{handle}" key="{key}"')

        return self._proposal_cache.setdefault(key, self._proposal(key))


    def __dir__(self):
        return super().__dir__()+[i for i in self.handles()]


    def __getattr__(self, handle):
        if handle not in self.handles():
            raise AttributeError(f'no such proposal: {handle}')        
        return self.proposal(handle=handle)


DataRepoBase = DataRepositoryBase

class DataScanBase:
    '''
    Abstract base class for `DataScan` objects.

    This is the main working unit for data preparation.
    A data scan is a series of data points ("frames" in
    `nx5d` lingo) collected in a single sweep or run of
    the measurement setup, for one particular scope.
    Generally, a "scan" has associated with it:
    
    - `.type`: a string that describes the type of scan,
      unambiguously within the framework of the data
      repository in use; scans of a type have the same
      data and positioners, and are to be analysed/processed
      in the same way. This is essential to determine the
      recipe for data processing.

    - `.summary`: a Key-Value store ("dictionary") of information
      that can be rapidly obtained about the scan without doing
      a full-open or full-retrieve of all the scan data.
      The specific keys are generally dependent on the exact
      `.type`, but they are all required to have at least
      a "description" field, which will be shown alongside
      name and kind in listings.

    - `.logbook`: a Markdown lab log book entry to accompany
      the scan

    - `.raw`: the collection of all data that was measured and
      is supposed to remain unchanged; typically, this is in
      some version of "device coordinate system" (angles,
      intensities, distances...). This is an `xarray.Dataset`.

    - `.cooked`: experimental data transformed into a
      physically more meaningful format; typically this involves
      coordinate transformation, interpolation, merging data
      points ("frames") into a larger whole. This also is an
      `xarray.Dataset`.
    
    - `.spice`: a collection of data ("spice data") required
      for processing ("cooking") the data from its raw form
      into a physically accessible form. This is typically a
      `nx5d.spice.SpiceView` object.
    '''

    def __init__(self, proposal, key):
        self._proposal = proposal
        self._scan_key = key
        self._summary = None

    def _get_summary(self):
        # Called to create a new summary object. Needs to implement
        # a mechanism to determine at least 'description' and 'type'.
        # Any other keys are optional, but will (eventually) be displayed
        # in human-readable listing forms occasionally.
        raise RuntimeError('not implemented')

    def _get_raw(self):
        # Loads raw data
        raise RuntimeError('not implemented')

    def _get_logentry(self):
        raise RuntimeError('not implemented')


    def _cook(self, data, **spice):
        '''
        Expected to return a cooked version of the data.

        This is what gets called under the hood when `.cooked`
        or similar functions are involved. This is intended for
        subclasses to (re)implement, not as a public use API.

        The default implementation checks for the existence of a
        spice type called "recipes". If it's not found, a `RuntimeError`
        is raised. If the "recipes" spice record is found, it is
        expected to be a dictionary with the following properties:
        - any key (except for `__default__`, see below) will be
          compared against this scan's type. If it matches, the data
          will be cooked by that recipe. The match is case-sensitive.
        - if no key is found by the name of this scan's type, but
          a `__default__` key exists, then the recipe of the default
          key is used.

        The preference is that recipes for every scan types are
        registered once, seeded in the beginning. Recipes for known
        types can be overridden according to spice rules; but spice
        rules also dictate that, once the "recipes" type is seeded,
        no new keys can be added or removed. This means that on new
        scan types can be implemented "on-the-fly". As a (deliberately
        ugly) workaround we honor the "__default__" key, to still
        make processing possible, but strongly encourage proper the
        beamline engineer to properly define their scan types instead
        of relying on later additions.

        For spice that is to be saved on-disk, the recipe must be defined
        as a string. The string has a URL-like format, pointing to a
        routine / callable with the signature
        `func(scan_type, input_data, **spice)` which must cook the data
        and return it -- see also: `nx5d.recipe:load_recipe()` for
        specific URL schemes.

        For purely in-memory spice types, the recipe can also be
        a reference to a Python callable.

        Args:
          data: input data
          **spice: spice keys and (view) values.

        Returns: the cooked data.
        '''
        if 'recipes' not in spice:
            raise RuntimeError('no recipes available and cooking not implemented')

        recipes = spice['recipes']
        try:
            rt = recipes[self.type]
        except KeyError:
            try: rt = recipes['__default__']
            except KeyError:
                raise RuntimeError(f'{self.type}: no recipe in '
                                   f'{spice["recipes"].keys} '
                                   f'and no __default__ mentioned')

        if isinstance(rt, str):
            rcall = self._load_recipe(rt)
        elif hasattr(rt, "__call__"):
            rcall = rt
        elif rt is None:
            rcall = lambda *a, **kw: None
        else:
            raise RuntimeError(f'unknown recipe spec: {rt}')

        return rcall(self.type, data, **spice)


    def _load_recipe(self, recipe_url):
        ## Loads the recipe from URL        
        import urllib
        from nx5d.recipe import load_recipe
        return load_recipe(recipe_url)


    @property
    def summary(self):
        if self._summary is None:
            try:
                self._summary = self._get_summary()
            except UnknownScanType:
                self._summary = {
                    'type': 'raw',
                    'descrption': 'raw'
                }
        return self._summary

    @property
    def url(self):
        return self._proposal.scan_url.format(scan=self.key)

    @property
    def key(self):
        return self._scan_key

    @property
    def type(self):
        return self.summary['type']

    @property
    def cooked(self):
        if not hasattr(self, '_cooked'):
            smap = {
                k:v.data() for k,v in self.spice.data.items()
            }
            self._cooked = self._cook(self.raw, **smap)
        return self._cooked

    @property
    def logbook(self):
        return self._get_logentry()

    @property
    def spice(self):
        if self._proposal.spice is not None:
            return self._proposal.spice.view(self._scan_key)


    @property
    def raw(self):
        if not hasattr(self, '_raw'):
            self._raw = self._get_raw()
        return self._raw

    def _repr_html_(self):
        return \
            '<table>'+\
              f'<tr><th>key</th><th>type</th>'+\
              ''.join([
                f'<th>{k}</th>' for k in
                filter(lambda x: x != 'type', self.summary.keys())
              ]) +\
              f'</tr>' +\
              f'<tr><td>{self.key}</td><td>{self.type}</td>'+''.join([
                f'<td>{v[1]}</td>' for v in
                filter(lambda x: x[0] != 'type', self.summary.items())
              ])+\
              f'</tr></table>'

    def __repr__(self):
        return f'{self.__class__.__name__}[{self.key}]'


class StackedDataScan(DataScanBase):
    '''
    Implements access to a series of similar scans

    Delegates all access ("raw", "cooked" etc) to the underlying 
    '''

    def __init__(self, proposal, scans, coords, stype, key=None):
        '''
        Args:
          proposal: The parent proposal
        
          scans: A list of "key" arguments to pass to the underlying
            scan class.

          coords: List of coodinate names to stack along. For each of
            the coordinates, a new dimension will be prepended to the
            input data. `None` or empty list also accepted.

          stype: Deviation of the scan type to report when asked for
            `.type`. This can contain the `{}` formatter to be replaced
            with the type of the original scans (which should be the
            same for all scans in the series)

          key: The key of this scan
        '''
        if key is None:
            key = '-'.join([scans[0], scans[-1]])
        super().__init__(proposal, key)
        self._scan_keylist = scans
        self._scan_coords = coords
        self._reported_type = stype

    
    def _get_summary(self):
        smry = {}
        for s in self._scan_keylist:
            scan = self._proposal._scan(key=s)
            smry.update(scan.summary)
            smry['type'] = self._reported_type.format(s)
        return smry


    def _get_raw(self):
        # Stacks all raw data
        
        subscans = {
            k:self._proposal._scan(k) for k in self._scan_keylist
        }

        from xarray import concat as xr_cat

        data_stack = None
        for tmp in subscans.values():
            scan_data = tmp.raw
            scan_data['scan'] = tmp.key
            if data_stack is None:
                data_stack = scan_data
            else:
                data_stack = xr_cat([data_stack, scan_data], dim='index')

        # by this point, the data is already stacked along its natural
        # 'index' dimension; make sure all the coordinates are maked as
        # such.
        if (self._scan_coords is not None) and len(self._scan_coords) > 0:
            for c in self._scan_coords[::-1]:
                data_stack = data_stack.set_coords(c)

        return data_stack


    def _get_logentry(self):
        raise RuntimeError('not implemented')



class DataProposalBase:
    '''
    Abstract base class for `DataProposal` type of objects.

    A "proposal" is a collection of data scans (sometimes
    called "runs") which belong together historically and,
    generally, by topic.

    Generally, spice management starts at proposal level,
    although spice _access_ happens one level below (scan
    level).

    A proposal has:
    
    - `.handle`: a string to refer to the proposal to, suitable
      of being converted in a Python symbol name

    - `.key`: a string that the proposal will be recognized at
      in all other instances, in particular (but not limited to)
      at the repository level.

    - `.all()`: a list of all the handles of all its subscans
    '''

    def __init__(self, repo, key, spice=None):
        '''
        Constructs a proposal access object.

        This is never used directly, there are several implementations (see
        `nx5d.repo.filesystem.py` and `nx5d.repo.astor.py` for examples).
        Typically, they will _also_ need other information, e.g. a base
        URL of where to get the data from. But in any case, they need a
        reference to the parent repository object (`repo`) and an
        identifyable information (`key`) of the current proposal.
        
        Args:
          repo: The parent repository object
          key: Identifying information for this proposal
          spice: A `SpiceProposal` or `SpiceCollection` object.
            Defaults to `None` if not specified; in that case, a
            spice proposal from the parent object (`repo`) will
            be requested for the same key.
        '''
        self._repo = repo
        self._proposal_key = key
        
        if spice is not None:
            self._spice = spiece
        else:
            self._my_spice_repo = self._spice_repo()
            self._spice = self._spice_repo().proposal(key=key)


    def _spice_repo(self, spice_url=None):
        '''
        Called once, during `.__init__()`, to obtain a spice repository
        for this proposal. `spice_url` is an explicit spice repository
        URL to use. Typically, `spice_url` is only then different from
        `None` when the upper layers (e.g. a `Repository` class) was
        initialized by the user with a specific spice repo.

        If `spice_url` is `None`, the environment variable `NX5D_SPICE_REPO`
        will be queried first. If it is an empty string or unset, the
        parent repository's `.spice_url` property will be queried.
        If all these still end up in an undefined spice repo, then a
        memory-only spice repository is created specifically for this
        proposal.

        This method can be overwritten by derived classes to perform
        additional spice operations upon spice repo initialization (e.g.
        checking for and/or seeding default spice types for this specific
        data proposal).
        '''
        from os import environ
        if (spice_url is None):
            if (self._repo.spice_url is None):
                spi_url = environ.get('NX5D_SPICE_REPO', '')
            else:
                spi_url = self._repo.spice_url

        if spi_url in (None, ''):
            # Constructing a memory spice repo
            from nx5d.spice import MemorySpiceRepository
            s = MemorySpiceRepository()
            s.new(self._proposal_key)
            return s
        else:
            from nx5d.spice import FsSpiceRepository
            return FsSpiceRepository(spi_url)


    #def location(self, **k):
    #    '''
    #    Returns the base part of the URL, with all relevant keys (e.g.
    #    "{propsal}") already substituted. This is useful for lower classes
    #    (e.g. DataScan) to be able to construct paths relative to the base
    #    proposal folder.
    #    '''
    #    raise RuntimeError('not implemented')


    @property
    def key(self):
        return self._proposal_key


    @property
    def handle(self):
        return self.all()[self._proposal_key]

    @property
    def spice(self):
        return self._spice
    
    def _scan_candidates(self):
        '''
        Returns a key-handle map for all scans in this proposal.
        This is where the actual reading of the scan list is supposed to take
        place, and needs to be reimplemented by derived classes.
        '''
        raise RuntimeError('not implemented')


    def all(self):
        '''
        Returns a key-handle map for all scans in this proposal.

        This is the high-level function that is part of the API, and implements some
        additional logic. One of the things it implements is the "stack" logic, whereby
        we can instruct Nx5d to load several consecutive scans and stack them on top
        of one another, forming a new type (built-in, spice-triggered call to .stack(),
        so to speak -- an "autostacking" of sorts).

        Autostacking is controlled by means of a spice type "stack", with the following
        schema:
        ```
           "stack": {
               "type": str | None,
               "options": { ... }
           }
        ```
        Herein the fields have the following meaning:
        - type: the type to assign to the new virtual, stacked DataScan. If the type is
          `None` or an empty string, the previous stack series is ended, and no
          current series starts. If the type is any other string (even if it's the
          same as a previous one), a new stack series with the corresponding type is
          started.
        - options: Key-Value assignment passed along as keyword agruments to `.stack()`.
          Most interesting will probably be the `coords` key, an optional, list of strings
          specifying which data keys to be promoted to coordinate status.

        Currently the data is always stacked along the first (i.e. "index") dimension,
        but this may change in the future.
        '''
        candidates = self._scan_candidates()
        stacks = self.spice.collection.anchors("stack")        
        if len(stacks) == 1: ## Easy way out, we don't have stack info
            return candidates

        bnd = [ anchor for anchor,spice in stacks.items() ]

        def _append_stack(scan_stack, scan_list, prev_type):
            # Helper to create a stacked-scan and append it to the "official"
            # scan list
            stack_handle = f'{next(iter(scan_stack))[1]}_stack'
            stack_key = f'{stack_handle}_{prev_type}'
            scan_list.append((stack_key, stack_handle))
            self._stacks[stack_key] = {
                'key': stack_key,
                'stype': prev_type,
                'scans': [k[0] for k in scan_stack]}
            self._stacks[stack_key].update(prev_params)

        sitr = iter(candidates.items())
        prev_type = None
        prev_params = {}
        scan_list = []
        self._stacks = {}
        for anchor,stackbranch in stacks.items():
            sti = stackbranch['stack']
            scan_stack = []

            if anchor not in (None, ''):
                before, after = before_and_after(lambda x: x[0] < anchor, sitr)
                if prev_type in (None, ''):
                    scan_list += [b for b in before]
                else:
                    scan_stack = [b for b in before]
                sitr = iter(after)

            if len(scan_stack) > 0:
                _append_stack(scan_stack, scan_list, prev_type)
                scan_stack = []

            prev_type = sti.type
            prev_params = sti.options if sti.options is not None else {}

        scan_list += [a for a in sitr]

        return dict(scan_list)


    def keys(self):
        return self.all().keys()


    def handles(self):
        return self.all().values()


    def recipe(self, stype):
        '''
        Expected to return a function able to process raw data of type `stype`.
        '''
        raise RuntimeError('not implemented')


    def _scan(self, key):
        '''
        Returns a `DataProposal` object for the proposal of this
        repo designated by `handle` or by `key`.
        '''
        raise RuntimeError('not implemented')


    def scan(self, handle=None, key=None):
        if key is None:
            key = {h:k for k,h in self.all().items()}[handle]

        if hasattr(self, "_stacks") and key in self._stacks:
            si = self._stacks[key]
            return self.stack(**si)

        return self._scan(key)


    def stack(self, stype, scans, key=None, coords=None):
        '''
        '''
        return StackedDataScan(proposal=self, scans=scans,
                               coords=coords, stype=stype,
                               key=key)


    def __dir__(self):
        return super().__dir__() + [i for i in self.handles()]


    def __getattr__(self, handle):
        if handle not in self.all().values():
            raise AttributeError(f'No such scan: {handle}')
        return self.scan(handle=handle)


    def _repr_html_(self):
        candidates = [self.scan(key=k) for k in self.keys()]

        desc = lambda x: x.summary['description'] \
            if len(x.summary['description']) > 0 \
               else ' '.join([
                       f'{k}={v}' for k,v in
                       filter(lambda x: x[0] not in \
                              ('description', 'type', 'intg', 'steps'),
                              x.summary.items())
               ])

        k2h = self.all()
        return \
            '<table><tr> '+\
            '<th>scan</th> '+\
            '<th>handle</th> '+\
            '<th>type</th> '+\
            '<th>integration</th> '+\
            '<th>steps</th> '+\
            '<th>notes</th> '+\
            '</tr>\n'+\
            '\n'.join([
                f'<tr> '
                f'<td style="text-align:left">{s.key}</td> '
                f'<td style="text-align:left">{k2h[s.key]}</td> '
                f'<td>{s.type}</td> '
                f'<td>{s.summary.get("intg", "")}</td> '
                f'<td>{s.summary.get("steps", "")}</td> '
                f'<td style="text-align:left">{desc(s)}</td> </tr>' \
                for s in candidates
            ])+\
            '</table>'



from nx5d.spice import FsSpiceRepository, MemorySpiceRepository, BranchView

import logging, json

logger = logging.getLogger("slim")

'''
Spice is always being managed at proposal level.

However, the "spice repository" is meant to be able to select a
specific spice (proposal subfolder) from a collection of proposals.

The general recipe when working with spice would be:

 - Define the spice location using the NX5D_SPICE_REPO env var
   (e.g. "..../{proposal}/spice", if your spice is in a subfolder
   of a proposal folder).

 - Make sure the _specific_ spice folder exists ("mkdir -p .../spice")

 - Use the SLIM CLI utility for spice management; try "slim ?" for help.
   Generally, what you want to do with SLIM is:

   - Check for help: "slim ?"

   - Check if SLIM sees your proposals: "slim proposals?"

   - List the spice for a specific proposal: "slim <proposal> list"
     Initially your proposal list will be empty (because there's nothing
     in the spice folder!)

   - Seed the spice for your speific proposal, by piping JSON data
     contents into SLIM: "cat data.json | slim <proposal> seed <handle>=<type> -"
     e.g. for KMC3 experimental info:
     "cat nx5d/examples/kmc3/exp_info_pilatus.json | \
        slim 252-cw32-13756-roessle seed exp_info=xrd/geometry -"

   - View spice from a specific anchor point (i.e. "scan" or "run"):
     "slim <proposal> view <anchor>" e.g.
     "slim 252-cw32-roessle-13756 view r0003"

   - Update spice / modify values:
     "slim <proposal> update <type>@<anchor> <key>=<value>".
     Note that updating happens not only at the specified anchor,
     but already where the spice type is anchored. For instance:
     if a spice named "offsets" is anchored at "r0001", but
     the update command refers to "offsets@r0023", the value of the
     parameter (e.g. "slim ... update offsets@r0023 theta=3.14")
     will be changed at "r0001", _not_ at "r0023". See also below.

   - Re-anchor a spice type:
     "slim <proposal> anchor <handle> <anchor>", e.g.
     "slim ... anchor offsets r0023"
  

'''

def _json_from_args(*args):
    # Transforms a list of arguments into a JSON object:
    #
    #   - if the first argument is a single '-',
    #     read JSON from stdin
    #
    #   - if an argument is key=value pair, we use that
    #     (parsing the 'value' string according to JSON rules,
    #     i.e. possibly nesting JSON data types)
    #
    #   - if artument is an existing file path, attempt to
    #     read JSON from the corresponding file
    #
    #   - otherwise collect arguments, concatenate them into
    #     a string, and parse the string argument as a
    #     JSON object

    if len(args) == 0:
        return {}

    if len(args) == 1 and args[0] == '-':
        return json.load(sys.stdin)

    import os, sys

    _xpl = {} ## explicit key=value entries in 'args'
    _ext = {} ## externally loaded
    _jstr = ""
    for a in args:
        s = a.split('=')
        if len(s) == 2:
            _xpl[s[0]] = json.loads(s[1])
        elif len(s) == 1:
            if os.path.exists(s[0]):
                with open(s[0]) as f:
                    _ext.update(json.load(f))
            else:
                _jstr += s[0]

    if len(_jstr) > 0:
        try:
            _ext.update(json.loads(_jstr))
        except json.decoder.JSONDecodeError as e:
            raise RuntimeError(f'{_jstr[0:5]}...{_jstr[-5:]}: '
                               f'neither an existent file nor a valid JSON string: '
                               f'{e}')

    if len(_xpl) > 0:
        _ext.update(_xpl)

    return _ext
            

def _branch_by_address(collection, addr_spec):
    # Returns a spice branch from a collection, using an
    # address specification form the command line.
    # The address specification "addr_spec" can be either
    # a UUID, or a <handle>@<anchor> spec.

    if '@' in addr_spec:
        t, a = addr_spec.split('@')
        view = collection.view(a, t)
        target = view.data[t]
    else:
        cand = collection.find(startswith=True, insensitive=True, uuid=addr_spec)
        if len(cand)==1:
            target = cand[0]
        else:
            target = cand

    if not isinstance(target, BranchView):
        print(f'bad branch for "{addr_spec}": {target}')
        return -1

    return target


def _check_dry(args):
    if "--dry" in args:
        args.remove('--dry')
        return True
    return False


class Application:
    '''
    Main application model for 'sm', the 'Spice Manager'.
    '''
    def __init__(self, args=None, env=None):
        from os import environ
        from sys import argv
        self._args = args if args is not None else argv
        self._env = env if env is not None else environ

        url = self._env.get('NX5D_SPICE_REPO',
                            self._env.get('NX5D_DATA_REPO',
                                          'file://.'))

        fsrepo = FsSpiceRepository(url)
        if self._args[-1].startswith('--dry'):
            self._args = self._args[:-1]
            logger.info(f'Dry-run requested; write operations in-memory only.')
            self._repo = MemorySpiceRepository(fsrepo)
        else:
            self._repo = fsrepo


    def _doctitle(self, cmd):
        return ''.join([
            x.replace('\n', '') for x in  \
            getattr(self, f'_cmd_{cmd}').__doc__.split('\n')[0:2]
        ])


    def _docdetail(self, cmd):
        return '\n'.join(
            getattr(self, f'_cmd_{cmd}').__doc__.split('\n')[2:]
        )


    def _all_commands(self):
        return [
            c[5:] for c in \
            filter(lambda x: x.startswith('_cmd_'), self.__dir__())
        ]

    def help_and_exit(self, args):
        docstr = lambda x: self._doctitle(x)
        cmd_text = "\n".join([
            f"  {c:10s} {docstr(c)}" \
            for c in self._all_commands()
        ])
        help_text = \
            f"Usage: {self._args[0]} <proposal> <command> [<opts>...]\n"\
            f"\n"\
            f"Available commands:\n"\
            f"{cmd_text}\n"\
            f"\n"\
            f"Try `{self._args[0]} ? <command>` for detailed help on commands."
        print(help_text)
        return -1


    def command_help_and_exit(self, rest):
        if len(rest) == 0:
            return self.help_and_exit(self._args)
        cmd = rest[0]
        help_text = \
            f"{cmd}: {self._doctitle(cmd)}\n"\
            f"{self._docdetail(cmd)}"
        print(help_text)
        return -1


    def list_proposals(self, args):
        # This is used mostly for autocomplete of the first parameter
        print('\n'.join(
            filter((lambda x: x.startswith(args[0])) \
                   if len(args)>0 else lambda x: x,
                   self._repo.all())
        ))


    def list_commands(self, args):
        print('\n'.join(
            filter((lambda x: x.startswith(args[0])) \
                   if len(args)>0 else lambda x: x,
            self._all_commands())
        ))


    def list_anchors(self, pobj, args):
        '''
        Lists all anchor points (scans/runs) in a specific proposal
        '''
        print('\n'.join(pobj.anchors().keys()))


    def version_and_exit(self):
        print(f'Spice List Manager (slim) @ nx5d v{myver.version}')
        return 0


    def run(self):
        # Wrapper for ._run() to actually generate screen output
        r = self._run()
        if (r is not None) and type(r) != int:
            print(r)

        

    def _run(self):
        # Actual command that executes the application according to
        # the command line arguments.
        #
        # This returns a Python object (in most cases a string or
        # Pandas data object) which is then printed on the screen.
        # See .run() for a conveninent wrapper.
        #
        # The separation is done for better unit testing.
        
        if len(self._args) <= 1:
            return self.version_and_exit()

        if len(self._args)>=2:
            try:
                k = self._args[1]
                proc = {
                    '?': self.command_help_and_exit,
                    'help': self.command_help_and_exit,
                    '-h': self.command_help_and_exit,
                    'proposals?': self.list_proposals,
                    'commands?': self.list_commands,
                }[self._args[1]]
                return proc(self._args[2:])                
            except KeyError:
                print('no', k)
                pass

        #try:
        proposal = self._repo.proposal(key=self._args[1])
        print(f'Proposal: {proposal.name}')
        #except AttributeError:
        #    print(f'No "{prop}" at "{self._repo.url_for(prop)}"')
        #    return -1

        try:
            action = self._args[2]
        except IndexError:
            self.list_anchors(pobj, self._args[1:])
            return -1

        proc = getattr(self, f"_cmd_{action}")

        if hasattr(proc, "__call__"):
            return proc(action, proposal, self._args[3:])


    def _rev_shortid(self, x):
        return x.meta('uuid')[:6]


    def _rev_conflict_list(self, rlist):
        return [
            self._rev_shortid(d) for d in \
            filter(lambda x: x.meta('revision') == rlist[0].meta('revision'), rlist)
        ]


    def _cmd_anchor(self, cmd, pobj, args):
        '''
        Anchors a spice branch at a new location

        Synopsis: anchor <anchor> <handle>

        Args:
        
          handle: handle (type) of the spice data
        
          anchor: new anchor point

        Returns: the UUID of the anchor entry, if successful.
        '''

        if len(args) != 2:
            print('Handle and anchor?')
            return -1

        anchor, handle = args
        try:
            n = pobj.anchor(anchor, handle)
        except Exception as e:
            print(e)
            return -1

        return n['uuid']
        

    def _cmd_seed(self, cmd, pobj, args):
        '''
        Create a spice branch from scratch
        
        The command will store the spice JSON file at the location of
        the configured repository.

        Synopsis: create <handle> -|<filename}|<p>=<v> [<p2>=<v2> ...]
        
        Args:
        
          handle: a Python-compatible string to identify
            the spice object by in autocomplete-sensitive environments
        
          p, v: key-value pairs representing the spice data,
            where "p" will always be interpreted as string, and "v"
            will be subject to an opportunistic cast (i.e. we try to
            convert it into "higher types" like int or float, but fall
            back to str if we fail).
        
          -: introduces an alternative way of specifying spice
            data; if this is the only parameter, spice data is read
            as a JSON package from stdin.

        Returns: the UUID of the seed entry, if successful.
        '''

        if len(args) < 2:
            print('Handle and values?')
            return -1

        handle = args[0]
        data = _json_from_args(*(args[1:]))
        try:
            n = pobj.seed(handle, data, idem=False)
        except Exception as e:
            print(e)
            return -1

        return n['uuid']


    def _cmd_update(self, cmd, pobj, args):
        '''
        Updates values for specified spice type, relative to anchor.

        If the latest available revision for the specified anchor is free
        of conflicts, it is used as a basis for a new revision. In the new
        revision, the fields specified are updated to the corresponding
        new values. A spice JSON file is sent to the repository to the same
        URL as the spice tree was read from.    

        Synopsis: update [--dry] <item> <field>=<value> [<field2>=<value> [...]],

        Args:

          item: a spice banch, designated either by its latest UUID,
            or by a <type>@<anchor> designation.

          field: name of the field to update

          value: the new value

        Returns: the UUID of the new entry, if successful.
        '''
        
        if len(args) == 0:
            print(f'{cmd}: item?')
            return -1

        try:
            if '@' in args[0]:
                # .update() receives (anchor, handle) as its address,
                # so we need to invert the order here
                addr = args[0].split('@')[::-1]
            else:
                # This is expected to be a UUID.
                addr = [uuid.UUID(args[0])]
            params = _json_from_args(*(args[1:]))
            n = pobj.update(*addr, payload=params)

        except Exception as e:
            print(e)
            return -1
        
        return n['uuid']
        

    def _cmd_list(self, cmd, pobj, args):
        '''
        List all anchor points and their associated spice handles

        Prints a list with spice handles as column heads, and anchor
        points as its index, will be displayed. The contents of the
        list are the first digits of single UUIDs if the respective
        revision is valid, or a list of UUIDs if there's a conflict.
        The listed UUIDs represent the spice entries in conflict.        
        
        Synopsis: list
        '''

        return pobj.ls()


    def _cmd_view(self, cmd, pobj, args):
        '''
        Lists the revision history for a specific anchor

        Synopsis: view <anchor> [<handle> [<handle> ...]]

        Args:
          anchor: Reference of a scan, representing the anchor point
            for which the view should be valid. This need not be the
            actual anchor point of a specific spice type -- if this
            is any other scan reference (even an unexistent one), the
            spice entries anchored directly above it will be considered.

          *handles: list of spice types to display; if the list is empty,
            all available types will be displayed.

        Returns: a table with anchor points, spice type, revision and
          a list of parameters will be displayed.
        '''
        if len(args)==0:
            print(f'{cmd}: anchor within "{pobj.name}"?')
            return -1

        anchor = args[0]
        handles = args[1:]

        try:
            view = pobj.view(anchor, *handles, ignore_conflict=True)
        except SpiceUnavailable as e:
            print(e)
            return -1

        return view._as_pandas_dframe().fillna('')


def main(args=None, env=None):

    from os import environ
    from sys import argv
    app = Application(args if args is not None else argv,
                      env if env is not None else environ)

    return app.run()




import time
import schema, json, uuid, functools, logging
from nx5d.repo.filesystem import FilesystemWalker
from nx5d.repo.filesystem import DataRepository as FsDataRepository
from nx5d.repo.base import DataRepoBase
from nx5d.xrd.udkm import opcast

import glob as glb

import nx5d._version as myver

logger = logging.getLogger(__name__)

spice_schema = schema.Schema({
    # Version of the format
    'version': 'kmc3/1a',

    # Optional (obsolete?); a type string (needs to be locally unique)
    schema.Optional('type'): schema.Optional(str),

    # A string suitable to serve as a Python name for data of this type
    'handle': str,

    # UUID of this particular entry.
    'uuid': schema.Or(str, uuid.UUID),

    # Incrementing counter; the higher number wins for this particular
    # anchor.
    'revision': int,

    # Scan reference from which on this update is valid
    'anchor': schema.Or(str, int),

    # Data dictionary (if None, this erases the whole chain
    # for this anchor)
    'data': schema.Or(None, dict)
})

# Validators for the 'data' payloads of various spice types
spice_types = {
    'straight': schema.Schema({ 'param': float}),
    'conflicted': schema.Schema({ 'value': float})
}

class SpiceBranch:
    '''
    Easy way to access fields of a single spice branch/type,
    for a specific anchor, sorted by revision.

    As per the data model, there will always be the "current version"
    of a view, if there currently is no revision conflict. Properties
    of the current version can be queried with the properties:
      - `.keys()  .alues()  .iems()` -- dict-like access
      - `.meta()` -- access to spice-object metadata
      - `.data()  .xr()` -- access to spice-type data (dict or xarray like)
      - `.uuid()  .revision()  .anchor()` -- commonly required spice
        metadata headers

    Some dynamic properties:
      - `.conflicted()` -- tells whether the view is clean or has a revision
        conflict
      - ...? 
    '''
    def __init__(self, *slist):
        '''
        Args:
          slist: list of spice objects (for the same type)
        '''

        self._dlist = sorted(slist, key=lambda x: x['revision'], reverse=True)


    def latest(self):
        '''
        Returns a list of items of the latest revision.
        The list should only have one item if there are no conflicts.
        '''
        ri = iter(self._dlist)
        latest = next(ri)
        candidates = [latest]
        while True:
            try:
                another = next(ri)
                if another['revision'] == latest['revision']:
                    candidates.append(another)
                if another['revision'] < latest['revision']:
                    raise StopIteration()
            except StopIteration:
                return candidates


    def history(self):
        '''
        Returns all spice object entries.
        '''
        return self._dlist


    def conflicted(self):
        '''
        Returns a list of conflicted items in the latest revision,
        or an empty list if there is no conflict.
        '''
        i = iter(self.latest())
        first = next(i)
        candidates = []
        while True:
            try:
                candidates.append(next(i))
            except StopIteration:
                if candidates: # not empty? -> conflicted!
                    candidates.append(first)
                break
        return candidates


    def obj(self):
        '''
        Returns the underlying JSON object representation if there are no
        conflicts; raises `SpiceRevisionConflict` otherwise.
        '''
        i = iter(self.latest())
        first = next(i)
        try:
            second = next(i)
            raise SpiceRevisionConflict(first, second)
        except StopIteration:
            return first


    def __dir__(self):
        return self.obj()['data'].keys()


    def __getattr__(self, k):
        try:
            return self.obj()['data'][k]
        except KeyError:
            raise AttributeError(k)


    def __getitem__(self, k):
       return self.obj()['data'][k]


    def keys(self):
        return self.obj()['data'].keys()


    def values(self):
        return self.obj()['data'].values()


    def items(self):
        return self.obj()['data'].items()


    def meta(self, k=None):
        '''
        Returns meta-data for the spice entry (i.e. information
        on fields that are _not_ part of the data).

        This is only usable if the branch is unconflicted.
        '''
        obj = self.obj()
        d = {
            k:obj[k] for k in filter(lambda x: x != 'data', obj.keys())
        }

        if k is not None:
            return d[k]
        
        return d


    def uuid(self):
        ''' Returns the UUID of the latest revision, if the branch is unconflicted. '''
        return self.meta('uuid')

    
    def revision(self):
        ''' Returns the revision of the latest revision. Also usable for conflicted branches. '''
        return next(iter(self._dlist))['revision']


    def anchor(self):
        ''' Returns the anchor. Also usable for conflicted branches. '''
        return next(iter(self._dlist))['anchor']


    def data(self):
        '''
        Returns a dictionary with all data fields, if the branch is unconflicted.
        '''
        return {k:v for k,v in self.obj()['data'].items()}


    def xr(self):
        '''
        Returns an xarray dataset with all data fields, if the branch is unconflicted.
        '''
        import xarray
        return xarray.Dataset(self.data())


    def make_update(self, **updates):
        '''
        Returns an updated spice entry.

        Essentially, this just copies everything from the underlying
        dataset, updates the UUID and revision, and updates the data
        fields.
        '''
        obj = next(iter(self._dlist))        
        dcopy = obj['data'].copy()
        
        for k,v in updates.items():
            if k not in dcopy:
                raise RuntimeError(f'msg="Key not available" key={k}')
            
        dcopy.update(**updates)
        n = {
            'version':  'kmc3/1a',
            'handle':   obj['handle'],
            'uuid':     uuid.uuid4(),
            'revision': obj['revision']+1,
            'anchor':   obj['anchor'],
            'data': dcopy
        }
        
        return spice_schema.validate(n)

    
    def __repr__(self):
        obj = next(iter(self._dlist))
        return self.__class__.__name__ \
            + f"<{obj['handle']}:{obj['revision']}"\
            + f"{'#' if self.conflicted() else ''}>"\
            + "(" \
            + ', '.join([
                f'{k}={v}' for k, v in obj['data'].items()
            ]) \
            + ")"

    def _repr_html_(self):
        obj = next(iter(self._dlist))
        return '<table><tr><th>Handle</th><th>Data</th></tr>'+\
            '\n'.join([
                f'<tr><td>{k}</td><td>{v}</td></td>' \
                for k,v in obj['data'].items()
            ]) + \
            '</table>'


BranchView = SpiceBranch
    

class SpiceView:
    '''
    Views a number of branches (for one specific anchor).

    Typically, this object will have attributes (see `.__dir__()`)
    with all branch names (i.e. "spice types"), and each attribute
    will hold a `BranchView` instance as value. If there's
    a revision conflict, then the value will be a tuple of
    `BranchViews` instead.
    '''
    def __init__(self, view_point, **branches):
        self._view = view_point
        self._data = branches
        Branch = lambda x: x if not isinstance(x, tuple) else x[0]

    @property
    def view_point(self):
        return

    @property
    def data(self):
        return self._data

    def handles(self):
        return self._data.keys()

    def __dir__(self):
        return self.handles()

    def __getattr__(self, handle):
        try:
            return self._data[handle]
        except KeyError:
            raise AttributeError(handle)


    def __repr__(self):
        from pprint import pformat
        return self.__class__.__name__+'('+pformat(self._data)+')'


    def _as_pandas_dframe(self):
        # Returns a Pandas DataFrame representation of this view
        view = self.data

        _view_item_dict = lambda v: { 
            'revision': v['revision'],
            'uuid':     v['uuid'][:6],
            'anchor':   v['anchor'],
            'data':     ' '.join([f'{f}={n}' for f,n in v['data'].items()])
        }

        import pandas
        out = pandas.DataFrame()
        for k,v in view.items():
            row_vals = {
                'viewpoint': self._view,
                'handle': k,
                'clean': not v.conflicted()
            }

            for vi in v.latest():
                row_vals.update(_view_item_dict(vi))
                out = pandas.concat([
                    out, pandas.DataFrame({
                        k: {row_vals['viewpoint']: v } \
                        for k, v in row_vals.items()
                    })
                ], ignore_index=True)

        if len(out)>0:
            return out.set_index('viewpoint')
        else:
            return out


    def _repr_html_(self):
        return self._as_pandas_dframe().fillna('')._repr_html_()


class SpiceRevisionConflict(RuntimeError):
    def __init__(self, candidate, other):
        self.canidate = candidate
        self.other = other
        super().__init__(f'msg="Revision conflict" '
                         f'handle={candidate["handle"]} '
                         f'candidate={candidate["uuid"]} '
                         f'conflicting={other["uuid"]}')

class SpiceUnavailable(RuntimeError):
    def __init__(self, msg=None, **params):
        self._params = params
        if msg is None:
            msg = \
                f'msg="Unavailable for '+\
                ' '.join([f'{k}="{v}"' for k,v in params.items()])
        super().__init__(msg)


def _rev_shortid(x):
    # returns a short-uuid variant for spice branch 'x'
    return x['uuid'][:6]


def _rev_conflict_list(rlist):
    # `rlist` is a spice branch (revision list)
    # returns a list of short-UUIDs that conflict with the
    # first (i.e. most recent) one
    return [
        _rev_shortid(d) for d in \
        filter(lambda x: x.meta('revision') == rlist[0].meta('revision'), rlist)
    ]
        

class SpiceCollection:
    '''
    In-memory representation all the available spice for a proposal.

    From here, views of branches can be created (i.e. collections of
    spice for a specific anchor, containing specific types) -- these are
    anchored at or above a specific frame.
    '''
    def __init__(self):
        # Data model here is:
        #  { type: { anchor: [revision] }  }
        self._data = { }


    def add(self, sdata):
        '''
        Adds a new update entry to the spice chain.
        '''
        d = spice_schema.validate(sdata)
        t_part = self._data.setdefault(d['handle'], {})
        a_part = t_part.setdefault(d['anchor'], [])
        a_part.append(d)


    def find(self, startswith=False, insensitive=False, **meta_args):
        '''
        Returns a list with BranchView objects that match criteria.

        Args:
          startswith: if set to `True`, when checking for matches
            of string values, only check if the spice value starts
            with the suppied search value. Otherwise check for a full
            match
          insensitive: if set to `True`, the match for strings will be
            performed in a case-insensitive way.
          **meta_args: keys and search values for meta fields to search
            in.

        Returns: a list with matching branch views, or an empty list
        if no result matches.
        '''
        ret = []
        for st,anclist in self._data.items():
            for anc,revlist in anclist.items():
                for rev in revlist:
                    for k,v in meta_args.items():
                        if k not in rev:
                            continue
                        r = rev[k]
                        v2 = v.lower() if isinstance(v, str) and insensitive else v
                        r2 = r.lower() if isinstance(r, str) and insensitive else r
                        if startswith:
                            if not r2.startswith(v2):
                                continue
                        else:
                            if not r2 == v2:
                                continue
                        ret.append(BranchView(rev))
        return ret


    def startswith(self, insensitive=True, **meta_args):
        '''
        Convenient wrapper for `.find()` with different defaults.
        '''
        return self.find(insensitive=insensitive, startswith=True, **meta_args)


    def handles(self, *handles):
        '''
        Returns the spice repo by spice handles as a dictionary
        { 'handle': { 'anchor': [revisions...] } }
        '''
        return {
            k:v for k,v in \
            (self._data.items() \
             if len(handles)==0 \
             else filter(lambda x: x[0] in handles, self._data.items()))
        }


    def anchors(self, *handles):
        '''
        Returns the spice repo by anchor points, as a dictionary
        like:
           {
              'anchor': { 'handle': [revisions, ... ]  },
              ...
           }
        '''
        ret = {}
        for stype,v in filter(lambda x: (x[0] in handles or len(handles)==0), self._data.items()):
            for sanchor,w in v.items():
                a = ret.setdefault(sanchor, {})
                a.setdefault(stype, SpiceBranch(*w))
            #    sorted_revisions = sorted(w, key=lambda x: x['revision'],
            #                              reverse=True)
            #    t = a.setdefault(stype, sorted_revisions)
        return { k:v for k,v in sorted(ret.items()) }


    def view(self, scan, *handles, ignore_conflict=False):
        '''
        Returns a SpiceView for the specified frame and types.

        If no types are requested, all available types are returned.
        There's always the latest revision that's returned for every
        frame/type combination. Conflicts in the latest revision
        are either reported (raising `SpiceRevisionConflict`) or
        ignored.

        Args:
            scan: the scan reference of the data point we want to
              collect spice for (i.e. "virtual anchor"). Set to
              `None` to view the seed.
        
            *handles: list of spice handles (strings)
            ignore_conflicts: if set to `False`, revision conflicts
              in the latest revision are ignored and all available
              revisions are returned.

        Returns: a `SpiceView` object.
        '''

        # BTW, the difference betwee a "scan" and an "anchor"
        # is this:
        #
        #  - a "scan" is (the reference of) a data collection run;
        #    it usually has a specific spice set associated with it,
        #    but mostly _indirectly_ through spice types published
        #    earlier
        #
        #  - an "anchor" is a scan which is boundary for the validity
        #    for a speific version of spice, which means that all previous
        #    scans don't have the (same version of) spice data for this
        #    type.

        if len(handles) == 0:
            search_handles = [k for k in  self._data.keys()]
        else:
            search_handles = handles

        branches = {}

        def revisions_for_anchor(data, viewpt):
            # Returns an (anchor-pt, revisions-list) for a specific
            # view point, or None if there's no revision in data.
            # `data` is for a specific type only.
            # If `viewpt` is `None`, return for the default anchor
            # (i.e. the seed).
            if (data is None) or  (len(data) == 0):
                raise RuntimeError(f'msg="No entries" stype={st}')

            # Sort by "anchor" (key of 'for_type') and find the one
            # that's immediately at or above requested frame.
            if viewpt is not None:
                tmp = sorted(data.items(), reverse=True)
                for a, rev_list in tmp:
                    if (a <= viewpt) and (len(rev_list) > 0):
                        return a, sorted(rev_list,
                                         key=lambda x: x['revision'],
                                         reverse=True)
            else:
                tmp = sorted(data.items(), reverse=False)
                if len(tmp) > 0:
                    rev_list = next(iter(tmp))[1]
                    return None, sorted(rev_list,
                                        key=lambda x: x['revision'],
                                        reverse=True)
                
            if len(data) != 0:
                msg = f'msg="Corrupt spice repo" detail="Unanchored entry"'
                logger.error(msg)
            raise SpiceUnavailable(msg=msg, anchor=viewpt)


        for st in search_handles:
            try:
                anc_pt, revisions = revisions_for_anchor(self._data.get(st, None), scan)
                tmp = branches[st] = SpiceBranch(*revisions)
                if not ignore_conflict:
                    chk = tmp.obj() ## will raise if there's a conflict
            except SpiceUnavailable as e:
                # If we have an explicit list with types, raise an error
                # if one is missing. Otherwise ignore.
                if handles is not None:
                    raise SpiceUnavailable(msg=str(e)+f' stype={st}', anchor=scan)

        return SpiceView(view_point=scan, **branches)


    def _list(self, *stypes):
        '''
        Creates a Pandas list with all data
        '''
        shortid = _rev_shortid
        conflict_list = _rev_conflict_list
        conflicted = lambda rlist: False if (len(rlist) <= 1) \
            else (rlist[0]['revision'] == rlist[1]['revision'])

        import pandas
        out = pandas.DataFrame()
        
        for anchor,branchlist in self.anchors(*stypes).items():
            cols = { 'anchor': {anchor: anchor}, }
            handles = {
                handle: {
                    anchor: shortid(branch.obj()) if not branch.conflicted() \
                    else ','.join([shortid(b) for b in branch.history()])
                } \
                for handle,branch in branchlist.items()
            }
            cols.update(handles)
            out = pandas.concat([out, pandas.DataFrame(cols)], ignore_index=True)

        if len(out)>0:
            return out.set_index('anchor').fillna('')
        else:
            return None

    
    def __repr__(self):
        return f'{self.__class__.__name__}[{",".join(self.handles().keys())}]'


    def _repr_html_(self):
        return self._list()._repr_html_()            


class SpiceProposal:
    '''
    Base class for access and storage-related operations for `SpiceCollection`

    Implements main operations like saving, seeding, updating etc.
    This is usually not used directly, it's subclassed. The subclass
    needs to implement at least the `.save()` and `.load_collection()`
    members.

    The most spice operations (seed, anchor, update) can be implemented
    generically as they're only logics around the actual collection object.
    '''
    def __init__(self, collection=None, name=None):
        '''
        Initializes the proposal backend.

        Args:
          collection: if specified, the spice data collection.
            If this is `None` (the default), then `.init_collection()`
            is called.
          name: Human-readable proposal name, used for output. Typically,
            this should be the same as the proposal key that was used to
            load the proposal.
        '''
        self.name = name
        self._collection = collection if collection is not None \
            else self.init_collection()


    def init_collection(self):
        '''
        Returns an initialized/popuplated spice collection from disk.

        The default implementation just returns an empty (but otherwise
        functional) `SpiceCollection`. You'd have to subclass and implement
        this for a spice proposal based on persistent storage.
        '''
        return SpiceCollection()


    def save(self, obj):
        '''
        Unconditionally saves a spice object.

        Saving is also performed non-idempotently, i.e. even if there's
        already a similar object existing. This can lead to conflicts or
        other kinds of logical spice system corruptions -- use only
        if you know what you're doing. Otherwise, use functions like
        `.seed()` or `.update()` instead, with `idem=True`.

        Args:
          obj: the full spice object (i.e. on-disk object, not just
            the data keys)

        Returns: a copy of the spice object, as written to disk.
        '''
        data = spice_schema.validate(obj)
        data.update({'uuid': str(data['uuid'])})
        self.collection.add(data)
        return data


    @property
    def collection(self):
        return self._collection


    def ls(self, *stypes):
        '''
        Lists the contents as an type/anchor point overview.
        '''
        return self.collection._list(*stypes)


    def view(self, *a, **kw):
        '''
        Convenience wrapper for the underlying `SpiceCollection.view()` method.
        '''
        return self.collection.view(*a, **kw)


    def seed(self, handle, payload=None, anchor=None, revision=1,
             idem=True, **extra_payload):
        '''
        Seeds (i.e. "creates") a new spice type in this proposal.
        
        This uses `.save()` under the hood, but wraps around some code
        to create  UUIDs, revisions and anchor from scratch, and check
        for already existing spice types.

        Args:
          handle: the handle of the spice
          payload: a dictionary with spice fields and values
          anchor: if specified, this will be used as the anchor seed.
            Typically this is `null` (JSON), or `None` (Python),
            which means that the backend will do its best to use
            a useful default that always sorts at the top of the
            list.
          revision: revision for the seed -- the default of `1` is
            a good choice.
          idem: if set to `True` (the default), the operation is
            idempotent. This means that when called repeatedly with
            the same handle, only the first call will have an effect.
            All subsequent other calls will have no effect on the
            repository and will just return the data already in store.
            But repated calls will _not_ fail. If set to `False`,
            attempts to re-seed with the same handle will raise a
            `RuntimeError`.
          **extra_payload: any other named arguments will be merged
            into `payload`. This is for convenience, so that
            both calls of the kind `.seed(payload={"param": ...})` and
            of the kind `.seed(param=...)`. Note that the extra
            payload parameters will overwrite original payload
            parameters, if specified repeatedly.

        Returns: a copy of the spice object that was saved, as
        returned by `.save()`.
        '''

        if payload is None:
            payload = {}
        payload.update(extra_payload)

        if handle in self._collection.handles():
            v = self._collection.view(None, handle)
            item = getattr(v, handle)
            if idem:
                #print('idem seed:', item.meta('uuid'))
                return item.obj()
            else:
                raise ValueError(f'{handle}: already exists as uuid={item.meta("uuid")}')

        obj = {
            'version': 'kmc3/1a',
            'type': handle,
            'handle': handle,
            'uuid': uuid.uuid4(),
            'revision': revision if revision is not None else 1,
            'anchor': anchor if anchor is not None else '',
            'data': payload
        }
        
        return self.save(obj)

    
    def anchor(self, anchor, handle, idem=True):
        '''
        Creates an anchor point for a spice type

        Args:
          anchor: anchor point, i.e. scan to which to anchor the
            spice type.
          handle: spice handle for which to create the anchor point.
            Spice type must already have been seeded. (See `.seed()`).
          idem: if set to `True` (the default), 

        Returns: the result of `.save()` which is a copy of the spice
        data that was saved.
        '''
        v = self._collection.view(anchor, handle)
        obj = getattr(v, handle)
        if (obj.meta('anchor') == anchor):
            if idem:
                return obj.obj()
            else:
                raise RuntimeError(f'{handle}: already anchored at {anchor}, '
                                   f'use update instead')
        
        nobj = {
            'version': 'kmc3/1a',
            'type': handle,
            'handle': handle,
            'uuid': uuid.uuid4(),
            'revision': obj.meta('revision')+1,
            'anchor': anchor,
            'data': obj.data()
        }

        return self.save(nobj)


    def update(self, *addr, payload=None, idem=True,
               **extra_payload):
        '''
        Updates an existing spice type with respect to a viewpoint.

        Args:
          addr: which spice to update. This can be either specified as
            a (viewpoint, handle) combination (in which case `addr` must
            be two arguments), or as a single UUID string (in which case
            the number of `addr` arguments is 1.)
          payload: `dict` with keys with values to update. Keys must
            match those in the seeded version of the spice type.
          idem: if set to `True` (default) the values will be checked
            against the last available revision of the spice type
            for the anchor. If the values match, there will be no
            update and the already available version will be returned
            instead.
          **extra_payload: keywords specified here will be used to
            update the payload dictionary before applying.

        Returns: the full resulting spice object after update.
        '''
        if payload is None:
            payload = {}
        payload.update(extra_payload)

        if len(addr)==2: ## (viewpoint, handle)
            found = getattr(self.collection.view(addr[0], addr[1]), addr[1])
            if not isinstance(found, BranchView):
                raise RuntimeError(f'bad branch for "{addr[1]}@{addr[0]}": {found}')

        elif len(addr)==1: ## UUID
            if isinstance(addr[0], uuid.UUID):
                s = self._collection.find(startswith=False, insensitive=True, uuid=str(addr[0]))
            else:
                s = self.collection.find(startswith=True, insensitive=True, uuid=str(addr[0]))
                
            if len(s) != 1:
                raise RuntimeError(f'bad branch for "{addr[0]}": {s}')
    
            found = next(iter(s))


        new = found.make_update(**payload)
        for k,v in found.data().items():
            if ((new['data'][k] != v) or (not idem)):
                return self.save(new)

        return found.obj()


    def __repr__(self):
        n = self.__class__.__name__
        return n


    def _repr_html_(self):
        tmp = self.ls()
        if tmp is not None:
            return tmp._repr_html_()
        else:
            return self.__repr__()+' (empty)'


class FsSpiceProposal(SpiceProposal):
    '''
    SpiceProposal implementation for a filesystem-like backend

    The spice objects are stored as files (or file-like) entries,
    one object per file. The target of the file is a URL format,
    implementing the 'file' mechanism, and possibly others like
    's3' in the future.
    '''
    def __init__(self, folder, glob=None, on_load_fail='warn', name=None):
        '''
        Initialization of the proposal access.

        (FIXME: specifying collection, AND key, AND url, feels like
        an awful lot of superfluous information...)

        Args:
          folder: Spice (proposal) folder to work on
          glob: Globbing pattern for the spice files within
            the folder. By default (i.e. when set to `None`),
            a string representation of UUIDs.
          on_load_fail: One of "warn", "ignore" or "raise" -- what
            to do when encountering a spice file that can't be
            loaded.
          name: Human-readablye name for nicer errors. We recommend
            you put the repo key here.
        '''

        self._folder = folder
        if glob is None:
            self._glob = \
                '[a-fA-F0-9]'*8 + '-' + \
                '[a-fA-F0-9]'*4 + '-' + \
                '[a-fA-F0-9]'*4 + '-' + \
                '[a-fA-F0-9]'*4 + '-' + \
                '[a-fA-F0-9]'*12
        else:
            self._glob = '*'

        self._on_load_fail = on_load_fail

        super().__init__(name=name)

        
    def init_collection(self):
        '''
        (Re)loads the spice collection from disk into memory
        '''
        col = SpiceCollection()
        for p in glb.glob(f'{self._folder}/{self._glob}'):
            with open(p, 'r') as f:
                try:
                    col.add(json.load(f))
                except Exception as e:
                    if self._on_load_fail in ('raise',):
                        raise e
                    elif self._on_load_fail in ('warn',):
                        logger.warning(str(e))

        return col


    def save(self, obj):
        '''
        Overriding `.save()` to actually store data on disk.
        '''
        import os        
        fpath = os.path.join(self._folder, str(obj['uuid']))
        os.makedirs(os.path.dirname(fpath), exist_ok=True)
        with open(fpath, 'x') as f:
            r = super().save(obj)
            json.dump(r, f)
            return r

    
    def __repr__(self):
        return f'{self.__class__.__name__}[{self._folder}]'


class FsSpiceRepository(FsDataRepository):
    '''
    On-disk access to spice

    The idea is that for each proposal we have a proposal-dependent
    base URL (possibly) containing raw data, and containing the spice
    collection for the proposal. This class manages access to the
    repository as a whole, and delegates proposal-specific work
    to the appropriate `SpiceProposal` implementation.
    '''

    def __init__(self, url, **repo_args):
        '''
        Args:
          url: URL of the spice repository (which may be the same as
            the data repository, if spice is stored on-disk alongside
            data). This should resolve to the per-proposal spice folders,
            not the data folders (e.g. "/mnt/repo/{proposal}/spice/" if
            your spice is in a "./spice/" subfolder of the corresponding
            proposal).
          **repo_args: keyword arguments to pass on to the
            `FsDataProposal` subclass. We recommend at least
            `glob` and `proposal_k2h` to be set.
        '''
        self._u = url
        super().__init__(url, **repo_args)


    def _proposal(self, key):
        '''
        Returns a `SpiceProposal` for the specified proposal key.
        '''
        scm, path = self._split_url(self._u.format(proposal=key))
        return {
            'file': FsSpiceProposal(path, name=key),
        }[scm]


    def _split_url(self, url):
        # return (scheme, path) from url
        parts = url.split('://')
        if len(parts) == 1:
            return ('file', parts[0])
        elif len(parts) == 2:
            return parts
        raise RuntimeError(f'don\'t understand {url}: {parts}')

    
class MemorySpiceRepository(DataRepoBase):
    '''
    Purely in-memory spice repository

    Inside Nx5d this is mostly used for unit testing, but it might have
    some use outside (e.g. for quick spice-based setups inside Jupyter
    Notebooks without access to permanent storage).

    Spice proposals are saved in an internal dictionary.

    As an option, this repository can be based off a template of another
    repository. This is such, this class is a good way of goofing around
    with spice of an existing repository without creating any damage
    (because nothing is saved).
    '''
    def __init__(self, template=None):
        '''
        Initializes the internal repo, possibly from a starting template

        Args:
          template: if not `None`, it is expected to be an instance of
            another (type of) spice repository which's contents will be
            used as a starting point. Loading of the other repo's data
            happens lazily, when `.proposal()` is called.
        '''
        super().__init__('')
        self._storage = {} ## key -> collection map
        self._template = template
        self._template_proposals = {} if template is None \
            else { k:k for k in template.all() }


    def all(self):
        keys = { k:k for k in self._storage.keys() }
        if self._template is not None:
            keys.update({k:k for k in self._template.all() })
        return keys


    def new(self, key):
        if (key in self._storage):
            return
        
        if (self._template is not None) and (key not in self._template.all()):
            return

        self._storage[key] = SpiceProposal(name=key)


    def _proposal(self, key):
        if (key not in self._storage) and (self._template is not None):
            self._storage[key] = self._template.proposal(handle=key)
        return SpiceProposal(self._storage[key].collection, name=key)



#!/usr/bin/python3

from numpy import array as np_array
import logging

'''
Contains useful tools for working HDF5-like data files (...as opposed to
the rest of the nx5d.h5like submodules, which mostly contain API elements
to *generate* HDF5-like data access.
'''

__all__ = [ "HdfCarver", "DatasetEmpty" ]

class DatasetEmpty(Exception):
    ''' Raised by HdfCarver.__call__() when a specific dataset turns out empty.

    This feature is actually dedicated to `nx5d.scan.ScanReader`, which in turn
    uses slicer generators which are sometimes open-ended (i.e. don't necessarily
    have a *known* maximum number of items to return). Since most indexable
    elements in Python (lists, numpy arrays, HDF5 datasets...) will happily 
    continue returning empty sets when sliced beyond their boundaries, the only
    way of knowing when to stop slicing (e.g. by `.streaks()` of `ScanReader`)
    is for `HdfCarver` to actually signalize the end of data.

    This is done by raising `DatasetEmpty`.
    '''
    pass


class HdfCarver:
    '''A fancy way of reading structured data from HDF5-like objects.
    
    Takes a dictionary which may contain HDF5 paths in the values
    and "restricts" the template by replacing the templated values
    by the data at the corresponding HDF5 path.

    Works recursively on the data, i.e. if the dicitonary contains
    sub-structured data, the structure is retained. All values that
    don't contain HDF5 references are returned verbatim.

    For instance, using this template:
    ```
    {
       "h5data": "@/path/to/dataset",
       "static": 3.14,
       "nested": {
          "static": [3, 4, 5],
          "h5data": "@/path/to/another/dataset",
       }
    }
    ```

    Would return something like this (assuming that "/path/to/dataset"
    and "/path/to/another/dataset" are arrays containing `[1, 2]` and `[3, 4]`,
    respectively):
    ```
    {
       "h5data": [1, 2],
       "static": 3.14,
       "nested": {
          "static": [3, 4, 5],
          "h5data": [3, 4],
       }
    }
    ```
    
    '''
    
    def __init__(self, template, accept=None, paths=None, passthrough=False):
        '''Initialises the carver.

        Args:
        
            template: the configuration template dictionaray

            accept: Places a whitelist restriction on the template
              keys to process and return. This is useful e.g. when
              template dictionaries are to be processed
              only partially. Note that unprocessed keys are _not_
              returned. Default is `None`, which disables the whitelist
              restriction.

            paths: If specified, it's expected to contain a
              for path format substitutes, i.e. this enables the template
              to contain references liek "@{key}", where `{key}` will
              be replaced with whatever is in `paths` at the corresponding
              dictionary slot.

            passthrough: If set to `True`, instead of returning the
              the actual data for HDF5 references, the HDF5-like Python
              container will be returned (e.g. the dataset HDF5 node object).
              This is useful for accessing metadata provided by HDF5 (e.g.
              dataset sizes and shapes) without actually reading out
              possibly large amounts of data. Default is `False`.
        '''
        
        self._tmpl = { k:v for k,v in template.items() }
        self._paths = {k:v for k,v in (paths or {}).items() }
        self._accepted = accept or [k for k in template.keys()]
        self._pass = passthrough


    def __value(self, h5like, value, fmtKeys, slicer=slice(None), **opts):
        '''
        This is the heart of the HDF5 address data lookup system: it checks
        the contents of `value` and returns the corresponding data as expected.
        The trick is in correctly determining the meaning of `value` here.

        The simples assumption here is that `value` is, indeed, a real piece
        of data (i.e. a number or an array with numbers), in which case it
        should be returned verbatim.

        But mostly this isn't the case. `value` might typically differ from
        a pure value in two ways (or any combination thereof):

          1. `value` is a string reference to a path within a HDF5-like
             object (file or node) -- here referenced by `h5like`.

          2. `value` is a container (tuple or dict) that we need to recursively
             descend into and look at each item individually.

        
        For the case of (1), the distinction is the following:
        
          - If `value` is a string that begins with `@`, then it is taken
            as a HDF5 address relative to `h5like`. The string is assumed
            to be a format (e.g. "contain {key}") where the keys to be
            substituted are taken from `fmtKeys`.

          - If `value` is a tuple which's first item is a string that begins
            with "@", and *all* other items are `slice()` objects, then
            the first item of the tuple is assumed to be the HDF5 path
            (as above), but instead of returning the value as-is, it is
            sliced using the provided slicer.

          - Everything else is returned as-is.


        We enter the case of (2) if neither of the above is the case, i.e.
        if `value` is either a dict, or another type of container (tuple
        with different items than (string, slice, ...),

          - If `value` is a tuple that doesn't fit the ("addr", slice()...)
            form, or is a dictionary, we need to recursively call
            `__valyue()` for each of the items.

        
        The function finally returns something that has the same
        structure as `value`, but eventual HDF5 paths are replaced by
        actual data as requested.

        Options:

          nodeOnly: do not return the actual data, just a reference to the
            HDF5 node(s)

          allowEmpty: allow the result of read operations to be an empty
            set (i.e. array of length 0). Otherwise a `DatasetEmpty`
            error is raised when there is no data.
        '''

        my_slicer = slicer

        # Check if value is a fancy ("addr", slice(), ...) tuple.
        # If it is, rewrite `value` to actually contain only the address,
        # and append the slicing part `value` to the slicer argument.
        if isinstance(value, tuple) and \
           len(value)>=2 and \
           (isinstance(value[0], str) and value[0][0] == '@') and \
           np_array([isinstance(v, slice) for v in value[1:]]).all():
            if isinstance(slicer, tuple):
                my_slicer = (*slicer, *value[1:])
            else:
                my_slicer = (slicer, *value[1:])
            value = value[0]


        if isinstance(value, str):
            # The main feature: loading data from HDF5 if we encounter an "@..." string.
            if (len(value) == 0) or (value[0] not in  ['@', '$']):
                return value

            # Only HDF5 addresses prefixed with '@' will be subject to slicing.
            # Prefix '$' will take the data verbatim, no slicing.
            doSlicing = (value[0] == '@')
            
            nodeAddrFmt = value[1:]
            
            try:
                nodeAddr = nodeAddrFmt.format(**fmtKeys)

                node = h5like[nodeAddr]

                logging.debug("Loading data from %s (only node: %r)" % \
                              (nodeAddr, opts.get('nodeOnly', False)))
                
                if opts.get('nodeOnly', False):
                    return node
                else:
                    try:
                        if doSlicing:
                            s = tuple(my_slicer[:len(node.shape)]) \
                                if isinstance(my_slicer, tuple) else my_slicer
                            data = node[s]
                            #print("Slicing %s: %r" % (value, data))
                        else:
                            data = node[()]
                            #print("Not slicing %s: %r" % (value, data))

                        if data.shape[0] == 0 and not opts.get('allowEmpty', False):
                            raise DatasetEmpty
                        
                        return data
                    
                    #except TypeError:  # ...not sure when this happens (?)
                    #    return node
                    except ValueError: # This happens when value is a scalar \
                           # and needs to be sliced by ()
                        return node[()]
                        

            except KeyError as e:
                logging.error("Path substitition error in '%s' for '%s': %s" \
                              % (nodeAddrFmt, value, str(e)))
                raise

            except DatasetEmpty:
                raise
                    
            except Exception as e:
                logging.error("%s: error reading node (%r)" % (nodeAddr, str(e)))
                raise


        # Descending into each of the elements if we encounter python basic
        # data types dict, list, tuple. For everything else, return verbatim.
            
        elif isinstance(value, dict):
            return dict({k:self.__value(h5like, v, fmtKeys, slicer, **opts) \
                         for k,v in value.items()})
        
        elif isinstance(value, tuple):
            return tuple([self.__value(h5like, v, fmtKeys, slicer, **opts) \
                          for v in value])

        elif isinstance(value, list):
            return [self.__value(h5like, v, fmtKeys, slicer, **opts) for v in value]
            
        else:
            return value

        
    def __call__(self, h5like=None, paths=None, slicer=slice(None), **opts):
        '''
        Uses the stored template to generate meaningful content based on the
        data from `h5like`. Typically, this involves going through all the
        templates' values and checking whether they consist of real data
        or H5 paths that need to be read. All paths are relative to `h5like`.

        If `nodeOnly` is set to `True`, the HDF5 nodes at the "nx5:" dataset
        paths is returned. Otherwise the `.value` property is read.

          - `slicer`: If specified, it is applied to any data that is to
            be retrieved from the HDF5-like node, insofar as the data shape
            matches the dimensionality of the slicer; if the slicer has
            too many components (e.g. a 3D-slicer on 1D data), the slicer
            is sliently truncated.        
        '''

        if h5like is None:
            return {k:v for k,v in self._tmpl.items()}

        combinedPaths = {k:v for k,v in self._paths.items()}
        combinedPaths.update((paths or {}))

        retr = {}
        for k,v in self._tmpl.items():
            if k in self._accepted:
                val = self.__value(h5like, v, fmtKeys=combinedPaths,
                                   slicer=slicer, **opts)
                retr[k] = val
                continue

            if self._pass:
                retr[k] = v

        return retr



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 17 13:38:50 2025

@author: aleks
"""
import nx5d.xrd.udkm as nxu
import pytest, os


@pytest.fixture
def data_base():
    return os.path.abspath(
        os.path.join(".", "tests", "test_data", "udkm",
                     "{proposal}", "{scan}")
    )


def test_load(data_base):
    fac = nxu.DatasetFactory()

    ds = fac(
            data_base.format(
                proposal=os.path.join("2025", "20250408"),
                scan="162828"
            )
    )
    assert 'images' in ds.data_vars
    for k in ds.data_vars:
        assert (ds[k].shape[0] == ds['images'].shape[0])


    ds = fac(
            data_base.format(
                proposal=os.path.join("2025", "20250424"),
                scan="150008"
            )
    )
    assert 'images' in ds.data_vars
    for k in ds.data_vars:
        assert (ds[k].shape[0] == ds['images'].shape[0])
        

def test_repo(data_base):
    print('Test repo:', data_base)
    repo = nxu.LegacyRepository(data_base)

    for pack in ('p20250424', 'p20250408'):
        assert pack in repo.__dir__()
        pobj = getattr(repo, pack)
        assert isinstance(pobj, nxu.DataPack)


def test_scan(data_base):
    print('Test repo:', data_base)
    repo = nxu.LegacyRepository(data_base)
    print('UDKM test repo:', repo.all())

    prop = repo.p20250424
    print('UDKM test proposal:', prop.all())

    html = prop._repr_html_()
    text = prop.__repr__()
    spice = prop.spice
    print('spice:', spice)

    scan = prop.r150008
    r = scan.raw
    assert "images" in r
    print(r)

    # 'prop' is now an in-memory proposal.
    #
    # It may or may not come with a recipe for the current scan type.
    # If it doesn't have, we seed one; if it does have recipes, we
    # make sure that at least the default fallback is something
    # we can use.
    if hasattr(scan.spice, "recipes"):
        prop.spice.update(None, 'recipes', __default__='pymod:///nx5d.recipe#my_callable')
    else:
        prop.spice.seed('recipes', **{ scan.type: 'pymod:///nx5d.recipe#my_callable'})
    
    c = scan.cooked
    print(c)



import logging, importlib, urllib

logger = logging.getLogger(__name__)

all = [
    "load_recipe",
    "pymod_recipe_loader",
    "file_recipe_loader"
]

def load_recipe(url):
    '''
    Loads a recipe from a URL-style location.

    The URL-like string supports the following schemes:
        - `pymod`: URL like "pymod://<netloc>/<module>#<callable>",
          where <netloc> is currently ignored, <module> needs to be
          a locally installed Python module in its dot-notation,
          and <callable> a callable inside that module.
        - `file`: URL like "file:///<path-to-file.py>#<callable>",
          loading <callable> from the local file <path-to-file.py>    
    '''
    
    purl = urllib.parse.urlparse(url)
        
    rloader = {
        'pymod': pymod_recipe_loader,
        'file': file_recipe_loader,
    }[purl.scheme]

    return rloader(purl)


def my_callable(stype, idata, **spice):
    # Just a test callable -- for unit testing, and for documentation
    return 'Jinkies!'

def pymod_recipe_loader(url):
    '''
    Loads a callable from a local python module.

    Expects the module name in the `.path` part of the (parsed) URL,
    and the callable inside the module. The initial "/" in the
    module path will be ignored. This is because of how URL definition
    (and parsing) works. Sub-modules can be separated either Python-style
    by ".", or URL-style by "/" (e.g. "module.submodule", "/module.submodule",
    "/module/submodule" will all point to the same module).
    The `.fragment` part (i.e. the part that follows after `#` in a URL) is
    the callable. Currently no other parameters are supported.

    This function doesn't check that the callable has a particular
    signature, but Nx5d's `DataScanBase` expects it to be of
    the form `proc(type, input, **spice)`.

    Returns a referene to the recipe callable.
    '''
    # Don't know how to handle net locations
    if url.netloc not in ('', None):
        raise RuntimeError(f'netloc must be empty (found: {url.netloc})')

    if url.path[0] == '/':
        modname = url.path[1:].replace('/', '.')
    else:
        modname = url.path.replace('/', '.')

    procparts = url.fragment.split('?')
    procname = procparts[0]
    if len(procparts) > 1:
        logger.warning('msg="URL parameters are ignored"')

    mobj = importlib.import_module(modname)
    return getattr(mobj, procname)
    

def file_recipe_loader(url):
    '''
    Loads a recipe from a file on the local filesystem.

    This works pretty much as expected: `.path` needs to point to the
    (absolute or relative) Python file name, and `.fragment` needs to
    be the callable inside the module

    Returns the recipe callable.
    '''
    
    import importlib.util
    import sys
    tmp = '_'.join(url.path.replace('.', '_').split('/'))
    modname = f'{__name__}.{tmp}'

    spec = importlib.util.spec_from_file_location(modname, url.path)
    mobj = importlib.util.module_from_spec(spec)

    procparts = url.fragment.split('?')
    procname = procparts[0]
    if len(procparts) > 1:
        logger.warning('msg="URL parameters are ignored"')

    sys.modules[modname] = mobj
    spec.loader.exec_module(mobj)

    return getattr(mobj, procname)
